import typing

import unicodedata
import html
from iso3166 import countries_by_alpha2
from pyasn1_alt_modules import rfc5280

from pkilint import validation, document
from pkilint.common import organization_id
from pkilint.common.organization_id import ParsedOrganizationIdentifier
from pkilint.itu import x520_name, asn1_util
from pkilint.pkix import Rfc2119Word


class ValidCountryCodeValidatorBase(validation.TypeMatchingValidator):
    def __init__(self, type_oid, value_path, checked_validation):
        super().__init__(
            type_path="type",
            type_oid=type_oid,
            value_path=value_path,
            pdu_class=rfc5280.AttributeTypeAndValue,
            validations=[checked_validation],
        )

    def validate_with_value(self, node, value_node):
        country_code = str(value_node.pdu)

        if country_code == "XX":
            return
        elif country_code not in countries_by_alpha2:
            raise validation.ValidationFindingEncountered(
                self.validations[0], f'Invalid country code: "{country_code}"'
            )


class ValidCountryValidator(ValidCountryCodeValidatorBase):
    VALIDATION_INVALID_COUNTRY_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.invalid_country_code"
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_at_countryName,
            value_path="value.x520countryName",
            checked_validation=self.VALIDATION_INVALID_COUNTRY_CODE,
        )


class CabfOrganizationIdentifierValidatorBase(
    organization_id.OrganizationIdentifierValidatorBase
):
    VALIDATION_ORGANIZATION_ID_INVALID_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.invalid_organization_identifier_registration_scheme",
    )

    VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.invalid_organization_identifier_country",
    )

    # the attribute name for this finding is prefixed with an underscore, so it's not flagged by the "validation report"
    # test
    _VALIDATION_ORGANIZATION_ID_INVALID_SP_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.invalid_organization_identifier_state_province_format",
    )

    REFERENCE_REQUIRED = (
        Rfc2119Word.MUST,
        "cabf.organization_identifier_reference_missing_for_scheme",
    )

    STATE_PROVINCE_PROHIBITED = (
        Rfc2119Word.MUST_NOT,
        "cabf.invalid_subject_organization_identifier_state_province_for_scheme",
    )

    _NTR_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_COUNTRY_CODES,
            VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
        ),
        state_province=(Rfc2119Word.MAY, None),
        reference=REFERENCE_REQUIRED,
    )

    _VAT_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_COUNTRY_CODES
            | {
                organization_id.COUNTRY_CODE_GREECE_TRADITIONAL,
                organization_id.COUNTRY_CODE_NORTHERN_IRELAND,
            },
            VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
        ),
        state_province=STATE_PROVINCE_PROHIBITED,
        reference=REFERENCE_REQUIRED,
    )

    _PSD_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_COUNTRY_CODES,
            VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
        ),
        state_province=STATE_PROVINCE_PROHIBITED,
        reference=REFERENCE_REQUIRED,
    )

    _ALLOWED_SCHEME_MAPPINGS = {
        "NTR": _NTR_SCHEME,
        "VAT": _VAT_SCHEME,
        "PSD": _PSD_SCHEME,
    }

    def __init__(
        self,
        invalid_format_validation: typing.Optional[validation.ValidationFinding],
        additional_schemes: typing.Optional[
            typing.Mapping[str, organization_id.OrganizationIdentifierElementAllowance]
        ] = None,
        enforce_strict_state_province_format=True,
        additional_validations=None,
        **kwargs,
    ):
        self._allowed_schemes = self._ALLOWED_SCHEME_MAPPINGS.copy()

        if additional_schemes:
            self._allowed_schemes.update(additional_schemes)

        self._enforce_strict_state_province_format = (
            enforce_strict_state_province_format
        )

        additional_validations = (
            [] if additional_validations is None else additional_validations.copy()
        )
        additional_validations.append(self.VALIDATION_ORGANIZATION_ID_INVALID_SCHEME)

        if self._enforce_strict_state_province_format:
            additional_validations.append(
                self._VALIDATION_ORGANIZATION_ID_INVALID_SP_FORMAT
            )

        super().__init__(
            self._allowed_schemes,
            invalid_format_validation,
            additional_validations=additional_validations,
            **kwargs,
        )

    @classmethod
    def handle_unknown_scheme(
        cls, node: document.PDUNode, parsed: ParsedOrganizationIdentifier
    ):
        raise validation.ValidationFindingEncountered(
            cls.VALIDATION_ORGANIZATION_ID_INVALID_SCHEME,
            f'Invalid registration scheme: "{parsed.scheme}"',
        )

    def validate_with_parsed_value(self, node, parsed):
        if (
            self._enforce_strict_state_province_format
            and parsed.state_province is not None
        ):
            if len(parsed.state_province) != 2 or not parsed.state_province.isalpha():
                raise validation.ValidationFindingEncountered(
                    self._VALIDATION_ORGANIZATION_ID_INVALID_SP_FORMAT,
                    f'State/province "{parsed.state_province}" is not two letters (will be fixed in erratum ballot)',
                )

        return super().validate_with_parsed_value(node, parsed)


class CabfOrganizationIdentifierAttributeValidator(
    CabfOrganizationIdentifierValidatorBase
):
    VALIDATION_ORGANIZATION_ID_INVALID_ENCODING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.invalid_subject_organization_identifier_encoding",
    )

    VALIDATION_ORGANIZATION_ID_INVALID_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.invalid_subject_organization_identifier_format",
    )

    def __init__(
        self,
        additional_schemes: typing.Optional[
            typing.Mapping[str, organization_id.OrganizationIdentifierElementAllowance]
        ] = None,
        enforce_strict_state_province_format=True,
        additional_validations: typing.Optional[
            typing.List[validation.ValidationFinding]
        ] = None,
    ):
        if additional_validations is None:
            additional_validations = []

        super().__init__(
            self.VALIDATION_ORGANIZATION_ID_INVALID_FORMAT,
            additional_schemes,
            enforce_strict_state_province_format,
            [self.VALIDATION_ORGANIZATION_ID_INVALID_ENCODING] + additional_validations,
            pdu_class=x520_name.X520OrganizationIdentifier,
        )

    @classmethod
    def parse_organization_id_node(
        cls, node
    ) -> organization_id.ParsedOrganizationIdentifier:
        name, value_node = node.child

        if name not in {"utf8String", "printableString"}:
            raise validation.ValidationFindingEncountered(
                cls.VALIDATION_ORGANIZATION_ID_INVALID_ENCODING,
                f"Invalid ASN.1 encoding: {name}",
            )

        return organization_id.parse_organization_identifier(str(value_node.pdu))


class RelativeDistinguishedNameContainsOneElementValidator(validation.Validator):
    VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE, "cabf.rdn_contains_multiple_atvs"
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS],
            pdu_class=rfc5280.RelativeDistinguishedName,
        )

    def validate(self, node):
        child_count = len(node.children)

        if child_count > 1:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS
            )


VALIDATION_INTERNAL_IP_ADDRESS = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR, "cabf.internal_ip_address"
)

VALIDATION_INTERNAL_DOMAIN_NAME = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR, "cabf.internal_domain_name"
)


class SignificantAttributeValueValidator(validation.Validator):
    VALIDATION_INSIGNIFICANT_ATTRIUBTE_VALUE_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.insignificant_attribute_value_present",
    )

    # https://www.unicode.org/reports/tr44/#General_Category_Values
    # TODO: any letter, number, or symbol is significant. revisit this to restrict S to only Sc (currency symbol)?
    _SIGNIFICANT_MAJOR_CLASSES = {
        "L",
        "N",
        "S",
    }

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_INSIGNIFICANT_ATTRIUBTE_VALUE_PRESENT],
            pdu_class=rfc5280.AttributeTypeAndValue,
        )

    def validate(self, node):
        value = asn1_util.get_string_value_from_attribute_node(node)

        if value is None:
            return

        if not any(
            unicodedata.category(c)[0] in self._SIGNIFICANT_MAJOR_CLASSES for c in value
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INSIGNIFICANT_ATTRIUBTE_VALUE_PRESENT,
                f'Insignificant attribute value: "{value}"',
            )


class HTMLEntitiesValidator(validation.Validator):
    """Validates that attribute values do not contain HTML entities using html.unescape."""

    VALIDATION_ATTRIBUTE_VALUE_CONTAINS_HTML_ENTITY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.name.attribute_value_contains_html_entity",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_ATTRIBUTE_VALUE_CONTAINS_HTML_ENTITY],
            pdu_class=rfc5280.AttributeTypeAndValue,
        )

    def validate(self, node):
        value_str = asn1_util.get_string_value_from_attribute_node(node)

        if not value_str:
            return

        unescaped_value = html.unescape(value_str)

        if value_str != unescaped_value:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ATTRIBUTE_VALUE_CONTAINS_HTML_ENTITY,
                f'Attribute value contains HTML entity: "{value_str}"',
            )



import csv
from os import path


def _test_csv(filename, expected_fieldnames):
    cur_dir = path.dirname(__file__)
    filename = path.join(cur_dir, "..", filename)

    with open(filename, "r", encoding="utf8") as csvfile:
        reader = csv.DictReader(csvfile, restkey="extra")

        assert reader.fieldnames == expected_fieldnames

        for row_idx, row in enumerate(reader):
            lineno = row_idx + 1 + 1  # 1-based index and header row

            extra = row.pop("extra", None)
            assert extra is None, f'Row "{row}" (line {lineno}) has extra fields'

            assert all(
                c is not None for c in row.values()
            ), f'Row "{row}" (line {lineno}) has a None value'


def test_cabf_smime_finding_metadata():
    _test_csv(
        "pkilint/cabf/smime/finding_metadata.csv",
        ["severity", "code", "source", "description"],
    )


def test_cabf_serverauth_finding_metadata():
    _test_csv(
        "pkilint/cabf/serverauth/finding_metadata.csv",
        ["severity", "code", "description"],
    )


def test_etsi_finding_metadata():
    _test_csv("pkilint/etsi/finding_metadata.csv", ["severity", "code", "description"])



import argparse
import datetime
import functools
from typing import Type

import dateutil.parser

from pkilint import validation, report, loader, document
from pkilint.pkix.certificate import certificate_validity
from pkilint.report import REPORT_FORMATS, report_wrapper


class SeverityThresholdAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if values == "ALL":
            severity = None
        elif values not in [s.name for s in validation.ValidationFindingSeverity]:
            raise argparse.ArgumentError(self, f'Invalid severity value: "{values}"')
        else:
            severity = validation.ValidationFindingSeverity[values]

        setattr(namespace, self.dest, severity)


def add_severity_arg(parser):
    parser.add_argument(
        "-s",
        "--severity",
        type=str.upper,
        default=validation.ValidationFindingSeverity.INFO,
        help="The finding severity threshold; findings with a lesser severity will not be reported.",
        action=SeverityThresholdAction,
        choices=[s.name for s in validation.ValidationFindingSeverity] + ["ALL"],
    )


class ReportFormatAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        report_generator_cls = REPORT_FORMATS[values]

        setattr(
            namespace,
            self.dest,
            functools.partial(report_wrapper, report_generator_cls),
        )


def add_report_format_arg(parser):
    parser.add_argument(
        "-f",
        "--format",
        type=str.upper,
        default=functools.partial(report_wrapper, report.ReportGeneratorPlaintext),
        help="The format in which results will be output.",
        choices=list(REPORT_FORMATS.keys()),
        action=ReportFormatAction,
    )


class DocumentFormatAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if values not in (f.name for f in loader.DocumentFormat):
            raise argparse.ArgumentError(self, f'Invalid document format: "{values}"')

        document_format = loader.DocumentFormat[values]

        setattr(namespace, self.dest, document_format)


def add_document_format_arg(parser):
    parser.add_argument(
        "--document-format",
        type=str.upper,
        default=loader.DocumentFormat.DETECT,
        help="The format of the specified input document(s).",
        choices=[f.name for f in loader.DocumentFormat],
        action=DocumentFormatAction,
    )


class ValidityPeriodStartAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def get_retriever_class(cls) -> Type[document.ValidityPeriodStartRetriever]:
        pass

    def __call__(self, parser, namespace, values, option_string=None):
        value_casefolded = values.casefold()

        if value_casefolded == "document".casefold():
            retriever_instance = self.get_retriever_class()()
        elif value_casefolded == "now".casefold():
            retriever_instance = document.StaticValidityPeriodStartRetriever(
                datetime.datetime.now(tz=datetime.timezone.utc)
            )
        else:
            try:
                dt = dateutil.parser.isoparse(values)

                retriever_instance = document.StaticValidityPeriodStartRetriever(dt)
            except ValueError:
                raise argparse.ArgumentError(
                    self, f'Invalid value for validity period start: "{values}"'
                )

        setattr(namespace, self.dest, retriever_instance)


class CertificateValidityPeriodStartAction(ValidityPeriodStartAction):
    @classmethod
    def get_retriever_class(cls):
        return certificate_validity.CertificateValidityPeriodStartRetriever


def add_validity_period_start_arg(action_cls: Type[ValidityPeriodStartAction], parser):
    parser.add_argument(
        "--validity-period-start",
        action=action_cls,
        default=action_cls.get_retriever_class()(),
        help="The start of the validity period that is compared to effective dates to determine applicability of "
        'validations. Acceptable values are "DOCUMENT" (use the validity period indicated in the document being '
        'validated, "NOW" (use the current time), or an ISO 8601-formatted date/time value.',
    )


add_certificate_validity_period_start_arg = functools.partial(
    add_validity_period_start_arg, CertificateValidityPeriodStartAction
)


def add_standard_args(parser):
    add_severity_arg(parser)
    add_report_format_arg(parser)
    add_document_format_arg(parser)


# This ensures that if a large (>255) number of findings are reported, we don't accidentally exit with an
# exit code of 0. This could happen if the number of findings is a multiple of 256.
def clamp_exit_code(exit_code):
    return min(exit_code, 255)



import base64
import tempfile

import pytest

from pkilint import loader, document

_CERT_B64 = """MIIGrzCCBJegAwIBAgIUYsQ+Fan+RfQ1ToEaA+PeZh43OTEwDQYJKoZIhvcNAQEL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"""


_CERT_WITH_TRAILER_B64 = """MIIGxDCCBKygAwIBAgIMS9Kjc1CWWKwCO6fGMA0GCSqGSIb3DQEBCwUAMEUxCzAJBgNVBAYTAlVT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"""


_OCSP_RESPONSE_B64 = """MIIDnwoBAKCCA5gwggOUBgkrBgEFBQcwAQEEggOFMIIDgTCBsKIWBBQK46D+ndQl
dpi163Lrygznvz318RgPMjAyNDA0MDIxMjM3NDdaMIGEMIGBMFkwDQYJYIZIAWUD
BAIBBQAEIDqZRndWgHOnB7/eUBhjReTNYTTbCF66odEEJfA7bwjqBCBHSmyjAfI9
yff3B4cE4cf1/JbnFnX27YguerZcP1hFQwIEAarwDYAAGA8yMDI0MDQwMzEyMzc0
N1qgERgPMjAyNDA0MTAxMjM3NDdaMAoGCCqGSM49BAMDA2kAMGYCMQDRmVmiIb4D
m9yEXiv2XtoeQi6ftpjLmlBqqRIi+3htfF/OyjdHnFuh38cQKYqqrWYCMQDKiPct
Vu7SQs587d2ZBEHQH20j5AFiGGsbI1b3+C9ZK6NIzgD6DnWlDwpSfilEarOgggJT
MIICTzCCAkswggGuoAMCAQICAQEwCgYIKoZIzj0EAwQwODELMAkGA1UEBhMCWFgx
FDASBgNVBAoMC0NlcnRzICdyIFVzMRMwEQYDVQQDDApJc3N1aW5nIENBMB4XDTI0
MDQwMjEyMzc0N1oXDTI1MDQwMjEyMzc0N1owPDELMAkGA1UEBhMCWFgxFDASBgNV
BAoMC0NlcnRzICdyIFVzMRcwFQYDVQQDDA5PQ1NQIFJlc3BvbmRlcjB2MBAGByqG
SM49AgEGBSuBBAAiA2IABFsJAbiFIyluuRnVD/oanLN0vE1AlYYoK/7KEbHZWtu1
RzSvVwv4K3IozyJrz0wl3bz+Oxo605Qw7/dj4daNLhUdkXILd5W1jaazRjlhOo+5
tajaSMZ0cRf5kZ6EJPN+yKOBhzCBhDAdBgNVHQ4EFgQUCuOg/p3UJXaYtety68oM
57899fEwHwYDVR0jBBgwFoAUjsIUCWB26pA46TmuG21SxBd9n74wDAYDVR0TAQH/
BAIwADAOBgNVHQ8BAf8EBAMCB4AwEwYDVR0lBAwwCgYIKwYBBQUHAwkwDwYJKwYB
BQUHMAEFBAIFADAKBggqhkjOPQQDBAOBigAwgYYCQRQqjNYKbGXHdGXfEVvB//i+
DiG02hraU9kGNKXeiQcPdZRajQsY/hdZPVyaykkAFVQGv29yWmTrEax+r4oZTtzG
AkFJCwtJpi7m00Qx9r/ugNWsnCFSiKUdxuvj7mg9lJtz0hexRJZKFODWJG5dUh//
Bc2w8vywgYYoduXu4QLcoP17CA=="""


_CRL_B64 = """MIIBzTCBtgIBATANBgkqhkiG9w0BAQsFADAiMQswCQYDVQQGEwJYWDETMBEGA1UE
CgwKQ1JMcyAnciBVcxcNMjQwMzI1MTg0NzAwWhcNMjQwNDAxMTg0NzAwWqBgMF4w
CgYDVR0UBAMCAQEwHwYDVR0jBBgwFoAU/NE0t8uklbG2WeoLBWIe6JqPtDowLwYD
VR0cAQH/BCUwI6AeoByGGmh0dHA6Ly9mb28uZXhhbXBsZS9jcmwuZGxshAH/MA0G
CSqGSIb3DQEBCwUAA4IBAQAN8oDSvWsg3JvUJ4MkXvczaFb72VH0J/VL5PV2cBSm
MfaVBKnUsNr1IcxT06KF8gNrDTpKqJ9fetO290swZfcPt9sEVUBVQUpdlQc3tya1
jYWmFkA3tkpqH5rBCQa3CBm1Cg8cbFBtwWgWr70NsVvfD6etjAEP9Ze+MSXnGV0p
w9EeOV07HnSD/PGQwqCiaSn5DdIDVoH8eFSGmgNLw+b4SwUjmz8PqsZwvHxJvleV
1D8cj7zdR4ywgRMjEfJZ8Bp+Tdu64Gv0doDS0iEJIshLHYkcW1okpq/tPm8kKAbD
reparePNQwhScVcDiSL73eEBIPokgG3QhohiucP5MeF1"""


def _make_pem(b64, pem_label):
    return f"-----BEGIN {pem_label}-----\n{b64}\n-----END {pem_label}-----"


def _load_and_compare(loader_func, expected_doc_cls, expected_substrate):
    loaded = loader_func()

    assert isinstance(loaded, expected_doc_cls)
    assert loaded.substrate == expected_substrate


def _test_loader_obj(loader_instance, doc_b64):
    doc_cls = loader_instance._document_cls
    label = loader_instance._document_pem_label

    doc_pem = _make_pem(doc_b64, label)
    doc_der = base64.b64decode(doc_b64)

    _load_and_compare(
        lambda: loader_instance.load_b64_document(doc_b64, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_pem_document(doc_pem, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_der_document(doc_der, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_document(doc_b64, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_document(doc_pem, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_document(doc_der, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_document_or_file(doc_b64, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_document_or_file(doc_pem, "test"), doc_cls, doc_der
    )
    _load_and_compare(
        lambda: loader_instance.load_document_or_file(doc_der, "test"), doc_cls, doc_der
    )

    # format-specific file load
    with tempfile.TemporaryFile("w+") as f:
        f.write(doc_b64)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_b64_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_b64.encode())

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_b64_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+") as f:
        f.write(doc_pem)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_pem_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_pem.encode())

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_pem_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_der)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_der_file(f, "test"), doc_cls, doc_der
        )

    # format-agnostic load
    with tempfile.TemporaryFile("w+") as f:
        f.write(doc_b64)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_b64.encode())

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+") as f:
        f.write(doc_pem)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_pem.encode())

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_der)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_file(f, "test"), doc_cls, doc_der
        )

    # format-agnostic file or document load
    with tempfile.TemporaryFile("w+") as f:
        f.write(doc_b64)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_document_or_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_b64.encode())

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_document_or_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+") as f:
        f.write(doc_pem)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_document_or_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_pem.encode())

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_document_or_file(f, "test"), doc_cls, doc_der
        )

    with tempfile.TemporaryFile("w+b") as f:
        f.write(doc_der)

        f.flush()
        f.seek(0)

        _load_and_compare(
            lambda: loader_instance.load_document_or_file(f, "test"), doc_cls, doc_der
        )


def test_certificate_loader():
    _test_loader_obj(loader._RFC5280_CERTIFICATE_LOADER_INSTANCE, _CERT_B64)


def test_crl_loader():
    _test_loader_obj(loader._RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE, _CRL_B64)


def test_ocsp_response_loader():
    _test_loader_obj(loader._RFC6960_OCSP_RESPONSE_LOADER_INSTANCE, _OCSP_RESPONSE_B64)


def test_load_cert_with_trailer():
    with pytest.raises(document.SubstrateDecodingFailedError):
        loader.load_der_certificate(
            base64.b64decode(_CERT_WITH_TRAILER_B64), "test", "test"
        )



import datetime
import functools
import logging
from typing import Set, Optional, List

from cryptography import x509, exceptions
from pyasn1.codec.der.encoder import encode
from pyasn1.type import univ
from pyasn1.type.base import Asn1Type
from pyasn1_alt_modules import rfc5280, rfc3739

from pkilint import validation, pkix, document
from pkilint.document import Document, ValueDecoder
from pkilint.pkix import (
    extension,
    time,
    name,
    create_name_validator_container,
    general_name,
    algorithm,
    key,
)
from pkilint.pkix.certificate import (
    certificate_validity,
    certificate_extension,
    certificate_validator,
    certificate_key,
    certificate_name,
    certificate_transparency,
)

logger = logging.getLogger(__name__)


class RFC5280Certificate(Document):
    def __init__(self, substrate_source, substrate, name=None, parent=None):
        super().__init__(
            rfc5280.Certificate(), substrate_source, substrate, name, parent
        )

    @property
    def not_before(self):
        try:
            return time.parse_time_node(
                self.root.navigate("tbsCertificate.validity.notBefore")
            )
        except ValueError:
            return pkix.MAXIMUM_TIME_DATETIME

    @property
    def not_after(self):
        try:
            return time.parse_time_node(
                self.root.navigate("tbsCertificate.validity.notAfter")
            )
        except ValueError:
            return pkix.MAXIMUM_TIME_DATETIME

    @property
    def validity_period(self) -> datetime.timedelta:
        return (self.not_after - self.not_before) + datetime.timedelta(seconds=1)

    def _decode_and_append_extension(
        self, ext_oid: univ.ObjectIdentifier, ext_asn1_spec: Asn1Type
    ) -> Optional[document.PDUNode]:
        ext_and_idx = self.get_extension_by_oid(ext_oid)

        if ext_and_idx is None:
            return None

        ext, _ = ext_and_idx
        ext_value = ext.children["extnValue"]

        try:
            return document.decode_substrate(
                self, ext_value.pdu.asOctets(), ext_asn1_spec, ext_value
            )
        except ValueError:
            # suppress decoding errors, which will be reported by DecodingValidator instances
            return None

    @functools.cached_property
    def is_ca(self) -> bool:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_basicConstraints, rfc5280.BasicConstraints()
        )

        return bool(decoded.navigate("cA").pdu) if decoded else False

    @functools.cached_property
    def extended_key_usages(self) -> Set[univ.ObjectIdentifier]:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_extKeyUsage, rfc5280.ExtKeyUsageSyntax()
        )

        return (
            {eku_node.pdu for eku_node in decoded.children.values()}
            if decoded
            else set()
        )

    @functools.cached_property
    def qualified_statement_ids(self) -> Set[univ.ObjectIdentifier]:
        decoded = self._decode_and_append_extension(
            rfc3739.id_pe_qcStatements, rfc3739.QCStatements()
        )

        return (
            {qs.children["statementId"].pdu for qs in decoded.children.values()}
            if decoded
            else set()
        )

    @functools.cached_property
    def cryptography_object(self):
        return x509.load_der_x509_certificate(self.substrate)

    @functools.cached_property
    def is_self_issued(self):
        issuer_node = self.root.navigate("tbsCertificate.issuer")
        subject_node = self.root.navigate("tbsCertificate.subject")

        return encode(issuer_node.pdu) == encode(subject_node.pdu)

    @functools.cached_property
    def public_key_object(self):
        return key.convert_spki_to_object(
            self.root.navigate("tbsCertificate.subjectPublicKeyInfo")
        )

    def is_signed_with_key(self, public_key):
        tbs_octets = self.cryptography_object.tbs_certificate_bytes
        signature_hash_alg = self.cryptography_object.signature_hash_algorithm
        signature_octets = self.cryptography_object.signature

        return key.verify_signature(
            public_key, tbs_octets, signature_octets, signature_hash_alg
        )

    @functools.cached_property
    def is_self_signed(self):
        if not self.is_self_issued:
            return False

        public_key = self.public_key_object

        # return False if the certificate certifies an unsupported public key type
        if public_key is None:
            return False

        try:
            self.is_signed_with_key(public_key)

        # gracefully handle unsupported signature algorithms
        except exceptions.UnsupportedAlgorithm:
            return False

    def get_extension_by_oid(self, oid):
        tbs_cert = self.root.children["tbsCertificate"]

        # ensure there's extensions
        if "extensions" not in tbs_cert.children:
            return None

        extensions_node = tbs_cert.children["extensions"]

        for ext_idx, extension_node in extensions_node.children.items():
            ext_oid = extension_node.children["extnID"].pdu

            if ext_oid == oid:
                return extension_node, int(ext_idx)

        return None

    def get_name_attributes_by_type(self, oid, name_path):
        name_node = self.root.navigate(name_path)

        return name.get_name_attributes_by_type(name_node, oid)

    def get_issuer_attributes_by_type(self, oid):
        return self.get_name_attributes_by_type(oid, "tbsCertificate.issuer")

    def get_subject_attributes_by_type(self, oid):
        return self.get_name_attributes_by_type(oid, "tbsCertificate.subject")

    @functools.cached_property
    def policy_oids(self) -> Set[univ.ObjectIdentifier]:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_certificatePolicies, rfc5280.CertificatePolicies()
        )

        return (
            {pi.children["policyIdentifier"].pdu for pi in decoded.children.values()}
            if decoded
            else set()
        )

    def get_san_general_names_by_type(
        self, general_name_type: general_name.GeneralNameTypeName
    ) -> List[document.PDUNode]:
        decoded = self._decode_and_append_extension(
            rfc5280.id_ce_subjectAltName, rfc5280.SubjectAltName()
        )

        return (
            []
            if decoded is None
            else [
                gn.children[general_name_type]
                for gn in decoded.children.values()
                if general_name_type in gn.children
            ]
        )


def create_spki_decoder(
    subject_public_key_type_mappings, subject_public_key_parameters_type_mappings
):
    subject_public_key_decoder = certificate_key.SubjectPublicKeyDecoder(
        type_mappings=subject_public_key_type_mappings
    )
    subject_public_key_parameters_decoder = (
        certificate_key.SubjectPublicKeyParametersDecoder(
            type_mappings=subject_public_key_parameters_type_mappings
        )
    )

    return validation.ValidatorContainer(
        validators=[
            certificate_key.SubjectPublicKeyDecodingValidator(
                decode_func=subject_public_key_decoder
            ),
            certificate_key.SubjectPublicKeyParametersDecodingValidator(
                decode_func=subject_public_key_parameters_decoder
            ),
        ],
        pdu_class=rfc5280.SubjectPublicKeyInfo,
    )


def create_policy_qualifier_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path="policyQualifierId",
        value_path="qualifier",
        type_mappings=type_mappings,
    )

    return validation.DecodingValidator(
        decode_func=decoder, pdu_class=rfc5280.PolicyQualifierInfo
    )


def create_other_name_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path="type-id", value_path="value", type_mappings=type_mappings
    )

    return validation.DecodingValidator(
        decode_func=decoder, pdu_class=rfc5280.AnotherName
    )


def create_qc_statements_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path="statementId", value_path="statementInfo", type_mappings=type_mappings
    )

    return validation.DecodingValidator(
        decode_func=decoder, pdu_class=rfc3739.QCStatement
    )


def create_issuer_validator_container(additional_validators=None, **kwargs):
    if additional_validators is None:
        additional_validators = []

    if len(kwargs) == 0:
        kwargs["path"] = "certificate.tbsCertificate.issuer"

    return create_name_validator_container(
        [
            name.EmptyNameValidator(),
            general_name.MailboxAddressSyntaxValidator(pdu_class=rfc5280.EmailAddress),
        ]
        + additional_validators,
        **kwargs
    )


def create_subject_validator_container(additional_validators=None, **kwargs):
    if additional_validators is None:
        additional_validators = []

    if len(kwargs) == 0:
        kwargs["path"] = "certificate.tbsCertificate.subject"

    return create_name_validator_container(
        [
            certificate_name.SubjectEmailAddressInSanValidator(),
            general_name.MailboxAddressSyntaxValidator(pdu_class=rfc5280.EmailAddress),
            name.DomainComponentValidDomainNameValidator(pdu_class=rfc5280.Name),
        ]
        + additional_validators,
        **kwargs
    )


def create_validity_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
            certificate_validity.CertificateSaneValidityPeriodValidator(),
            time.TimeCorrectEncodingValidator(),
        ]
        + additional_validators,
        path="certificate.tbsCertificate.validity",
    )


def create_spki_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []

    return validation.ValidatorContainer(
        validators=[
            certificate_key.CaPrehashPublicKeyValidator(),
            certificate_key.ObsoletePublicKeyAlgorithmValidator(),
        ]
        + additional_validators,
        path="certificate.tbsCertificate.subjectPublicKeyInfo",
    )


def create_extensions_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
            extension.UniqueExtensionValidator(),
            certificate_extension.BasicConstraintsValidator(),
            certificate_extension.CertificatePolicyQualifierValidator(),
            certificate_extension.CertificatePoliciesUserNoticeValidator(),
            certificate_key.SubjectKeyIdentifierValidator(),
            certificate_extension.SubjectKeyIdentifierCriticalityValidator(),
            certificate_extension.KeyUsageCriticalityValidator(),
            certificate_extension.KeyUsageValidator(),
            certificate_key.SpkiKeyUsageConsistencyValidator(),
            general_name.UriSyntaxValidator(pdu_class=rfc5280.CPSuri),
            general_name.GeneralNameValidatorContainer(),
            certificate_extension.DuplicatePolicyValidator(),
            certificate_extension.CrlDpCriticalityValidator(),
            certificate_extension.NameConstraintsCriticalityValidator(),
            extension.AuthorityKeyIdentifierValidator(),
            extension.AuthorityKeyIdentifierCriticalityValidator(),
            certificate_extension.NameConstraintsValidator(),
            certificate_extension.NameConstraintsGeneralSubtreeValidator(),
            certificate_extension.AuthorityInformationAccessCriticalityValidator(),
            certificate_extension.SubjectInformationAccessCriticalityValidator(),
            certificate_extension.SubjectAlternativeNameCriticalityValidator(),
            certificate_extension.SubjectDirectoryAttributesCriticalityValidator(),
            certificate_extension.SmimeCapabilitiesCriticalityValidator(),
            extension.DistributionPointValidator(),
            certificate_extension.CtPrecertPoisonCriticalityValidator(),
            certificate_extension.CtPrecertPoisonSctListMutuallyExclusiveExtensionsValidator(),
            certificate_transparency.SctListElementCountValidator(),
            certificate_extension.PolicyConstraintsPresenceValidator(),
            certificate_extension.PolicyMappingsPresenceValidator(),
            certificate_extension.InhibitAnyPolicyPresenceValidator(),
            certificate_extension.ProhibitedQualifiedStatementValidator(),
            certificate_extension.IssuerAlternativeNameCriticalityValidator(),
            certificate_extension.ExtendedKeyUsageCriticalityValidator(),
        ]
        + additional_validators,
        path="certificate.tbsCertificate.extensions",
    )


def create_pkix_certificate_validator_container(decoding_validators, validators):
    decoding_validator_container = [
        validation.ValidatorContainer(
            validators=decoding_validators, path="certificate"
        )
    ]

    validators += [
        certificate_validator.CorrectVersionValidator(),
        pkix.CertificateSerialNumberValidator(),
        certificate_validator.SignatureAlgorithmMatchValidator(),
        certificate_extension.SubjectKeyIdentifierPresenceValidator(),
        certificate_extension.AuthorityKeyIdentifierPresenceValidator(),
        certificate_extension.KeyUsagePresenceValidator(),
        time.UtcTimeCorrectSyntaxValidator(),
        time.GeneralizedTimeCorrectSyntaxValidator(),
        certificate_validator.IssuerUniqueIdAbsenceValidator(),
        certificate_validator.SubjectUniqueIdAbsenceValidator(),
    ]

    return validation.ValidatorContainer(
        validators=decoding_validator_container + validators
    )


def create_decoding_validators(
    name_mappings, extension_mappings, additional_decoding_validators=None
):
    if additional_decoding_validators is None:
        additional_decoding_validators = []
    return [
        pkix.create_attribute_decoder(name_mappings),
        pkix.create_extension_decoder(extension_mappings),
        pkix.create_signature_algorithm_identifier_decoder(
            algorithm.SIGNATURE_ALGORITHM_IDENTIFIER_MAPPINGS,
            path="certificate.tbsCertificate.signature",
        ),
        create_spki_decoder(
            key.SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS,
            key.SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS,
        ),
        create_policy_qualifier_decoder(
            certificate_extension.CERTIFICATE_POLICY_QUALIFIER_MAPPINGS
        ),
        create_other_name_decoder(general_name.OTHER_NAME_MAPPINGS),
        certificate_transparency.SctListExtensionDecodingValidator(),
    ] + additional_decoding_validators



import enum
from enum import auto

from pyasn1.type.univ import ObjectIdentifier

BR_VERSION = "2.1.5"


ID_POLICY_EV = ObjectIdentifier("2.23.140.1.1")
ID_POLICY_DV = ObjectIdentifier("2.23.140.1.2.1")
ID_POLICY_OV = ObjectIdentifier("2.23.140.1.2.2")
ID_POLICY_IV = ObjectIdentifier("2.23.140.1.2.3")

SERVERAUTH_RESERVED_POLICY_OIDS = {
    ID_POLICY_EV,
    ID_POLICY_DV,
    ID_POLICY_OV,
    ID_POLICY_IV,
}


@enum.unique
class CertificateType(enum.IntEnum):
    ROOT_CA = auto()
    INTERNAL_CROSS_CA = auto()
    EXTERNAL_CROSS_CA = auto()
    INTERNAL_SUBSCRIBER_ISSUING_CROSS_CA = auto()
    EXTERNAL_SUBSCRIBER_ISSUING_CROSS_CA = auto()
    NON_TLS_CA = auto()
    PRECERT_SIGNING_CA = auto()
    INTERNAL_UNCONSTRAINED_TLS_CA = auto()
    INTERNAL_CONSTRAINED_TLS_CA = auto()
    EXTERNAL_UNCONSTRAINED_TLS_CA = auto()
    EXTERNAL_UNCONSTRAINED_EV_TLS_CA = auto()
    EXTERNAL_CONSTRAINED_TLS_CA = auto()
    EXTERNAL_CONSTRAINED_EV_TLS_CA = auto()
    DV_FINAL_CERTIFICATE = auto()
    IV_FINAL_CERTIFICATE = auto()
    OV_FINAL_CERTIFICATE = auto()
    EV_FINAL_CERTIFICATE = auto()
    OCSP_RESPONDER = auto()
    DV_PRE_CERTIFICATE = auto()
    IV_PRE_CERTIFICATE = auto()
    OV_PRE_CERTIFICATE = auto()
    EV_PRE_CERTIFICATE = auto()

    def __str__(self):
        return self.name

    @property
    def to_option_str(self):
        return self.name.replace("_", "-")

    @staticmethod
    def from_option_str(value):
        value = value.replace("-", "_").upper()

        return CertificateType[value]


INTERNAL_CROSS_CA_TYPES = {
    CertificateType.INTERNAL_CROSS_CA,
    CertificateType.INTERNAL_SUBSCRIBER_ISSUING_CROSS_CA,
}


EXTERNAL_CROSS_CA_TYPES = {
    CertificateType.EXTERNAL_CROSS_CA,
    CertificateType.EXTERNAL_SUBSCRIBER_ISSUING_CROSS_CA,
}


CROSS_CA_TYPES = INTERNAL_CROSS_CA_TYPES | EXTERNAL_CROSS_CA_TYPES


ROOT_KEY_CROSS_CA_TYPES = {
    CertificateType.INTERNAL_CROSS_CA,
    CertificateType.EXTERNAL_CROSS_CA,
}


INTERNAL_CA_TYPES = {
    CertificateType.INTERNAL_CROSS_CA,
    CertificateType.INTERNAL_SUBSCRIBER_ISSUING_CROSS_CA,
    CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA,
    CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.NON_TLS_CA,
}

EXTERNAL_CA_TYPES = {
    CertificateType.EXTERNAL_UNCONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_UNCONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_CROSS_CA,
    CertificateType.EXTERNAL_SUBSCRIBER_ISSUING_CROSS_CA,
}

INTERMEDIATE_CERTIFICATE_TYPES = (
    INTERNAL_CA_TYPES | EXTERNAL_CA_TYPES | {CertificateType.PRECERT_SIGNING_CA}
)

CONSTRAINED_TLS_CA_TYPES = {
    CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
}

TLS_CA_TYPES = (
    {
        CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA,
        CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
    }
    | EXTERNAL_CA_TYPES
    | INTERNAL_CROSS_CA_TYPES
)

SUBSCRIBER_FINAL_CERTIFICATE_TYPES = {
    CertificateType.DV_FINAL_CERTIFICATE,
    CertificateType.IV_FINAL_CERTIFICATE,
    CertificateType.OV_FINAL_CERTIFICATE,
    CertificateType.EV_FINAL_CERTIFICATE,
}

SUBSCRIBER_PRECERT_TYPES = {
    CertificateType.DV_PRE_CERTIFICATE,
    CertificateType.IV_PRE_CERTIFICATE,
    CertificateType.OV_PRE_CERTIFICATE,
    CertificateType.EV_PRE_CERTIFICATE,
}

SUBSCRIBER_CERTIFICATE_TYPES = (
    SUBSCRIBER_FINAL_CERTIFICATE_TYPES | SUBSCRIBER_PRECERT_TYPES
)

DV_CERTIFICATE_TYPES = {
    CertificateType.DV_PRE_CERTIFICATE,
    CertificateType.DV_FINAL_CERTIFICATE,
}
IV_CERTIFICATE_TYPES = {
    CertificateType.IV_PRE_CERTIFICATE,
    CertificateType.IV_FINAL_CERTIFICATE,
}
OV_CERTIFICATE_TYPES = {
    CertificateType.OV_PRE_CERTIFICATE,
    CertificateType.OV_FINAL_CERTIFICATE,
}
EV_CERTIFICATE_TYPES = {
    CertificateType.EV_PRE_CERTIFICATE,
    CertificateType.EV_FINAL_CERTIFICATE,
}

IDENTITY_CERTIFICATE_TYPES = (
    IV_CERTIFICATE_TYPES | OV_CERTIFICATE_TYPES | EV_CERTIFICATE_TYPES
)



from pyasn1_alt_modules import rfc3279, rfc5480, rfc5280

from pkilint.pkix import key
from pkilint import validation


class RsaKeyValidator(validation.Validator):
    VALIDATION_RSA_MODULUS_INVALID_LENGTH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.rsa_modulus_invalid_length"
    )

    VALIDATION_RSA_MODULUS_HAS_SMALL_PRIME_FACTOR = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.rsa_modulus_has_small_prime_factor",
    )

    VALIDATION_RSA_EXPONENT_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.rsa_exponent_prohibited_value"
    )

    VALIDATION_RSA_EXPONENT_NOT_IN_RECOMMENDED_RANGE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.rsa_exponent_not_in_recommended_range",
    )

    _PRIMES_UNDER_752 = {
        2,
        3,
        5,
        7,
        11,
        13,
        17,
        19,
        23,
        29,
        31,
        37,
        41,
        43,
        47,
        53,
        59,
        61,
        67,
        71,
        73,
        79,
        83,
        89,
        97,
        101,
        103,
        107,
        109,
        113,
        127,
        131,
        137,
        139,
        149,
        151,
        157,
        163,
        167,
        173,
        179,
        181,
        191,
        193,
        197,
        199,
        211,
        223,
        227,
        229,
        233,
        239,
        241,
        251,
        257,
        263,
        269,
        271,
        277,
        281,
        283,
        293,
        307,
        311,
        313,
        317,
        331,
        337,
        347,
        349,
        353,
        359,
        367,
        373,
        379,
        383,
        389,
        397,
        401,
        409,
        419,
        421,
        431,
        433,
        439,
        443,
        449,
        457,
        461,
        463,
        467,
        479,
        487,
        491,
        499,
        503,
        509,
        521,
        523,
        541,
        547,
        557,
        563,
        569,
        571,
        577,
        587,
        593,
        599,
        601,
        607,
        613,
        617,
        619,
        631,
        641,
        643,
        647,
        653,
        659,
        661,
        673,
        677,
        683,
        691,
        701,
        709,
        719,
        727,
        733,
        739,
        743,
        751,
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_RSA_MODULUS_INVALID_LENGTH,
                self.VALIDATION_RSA_MODULUS_HAS_SMALL_PRIME_FACTOR,
                self.VALIDATION_RSA_EXPONENT_PROHIBITED,
                self.VALIDATION_RSA_EXPONENT_NOT_IN_RECOMMENDED_RANGE,
            ],
            pdu_class=rfc3279.RSAPublicKey,
        )

    def validate(self, node):
        modulus = int(node.children["modulus"].pdu)
        modulus_len = modulus.bit_length()
        exponent = int(node.children["publicExponent"].pdu)

        results = []

        if modulus_len < 2048 or modulus_len % 8 != 0:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_MODULUS_INVALID_LENGTH,
                    f"Invalid modulus length: {modulus_len}",
                )
            )

        small_prime = next(
            filter(lambda p: modulus % p == 0, self._PRIMES_UNDER_752), None
        )

        if small_prime is not None:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_MODULUS_HAS_SMALL_PRIME_FACTOR,
                    f"Modulus has prime factor less than 752: {small_prime}",
                )
            )

        if exponent % 2 == 0:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_EXPONENT_PROHIBITED,
                    f"Exponent is not odd: {exponent}",
                )
            )
        elif exponent < 3:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_EXPONENT_PROHIBITED,
                    f"Exponent is less than 3: {exponent}",
                )
            )

        if exponent < ((1 << 16) + 1) or exponent > ((1 << 256) - 1):
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_EXPONENT_NOT_IN_RECOMMENDED_RANGE,
                    f"Exponent is out of recommended range: {exponent}",
                )
            )

        return validation.ValidationResult(self, node, results)


class EcdsaKeyValidator(validation.Validator):
    VALIDATION_KEY_VALIDATION_FAILED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING, "cabf.ecdsa_key_validation_failed"
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_KEY_VALIDATION_FAILED],
            pdu_class=rfc5280.SubjectPublicKeyInfo,
            predicate=lambda n: n.navigate("algorithm.algorithm").pdu
            == rfc5480.id_ecPublicKey,
        )

    def validate(self, node):
        try:
            key_obj = key.convert_spki_to_object(node)

            if key_obj is None:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_KEY_VALIDATION_FAILED, "Unsupported key type"
                )
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_KEY_VALIDATION_FAILED, str(e)
            )



import re
import typing
from typing import List

from pyasn1.type import univ
from pyasn1_alt_modules import rfc5280, rfc3739

from pkilint import validation, document
from pkilint.common import organization_id
from pkilint.common.organization_id import ParsedOrganizationIdentifier
from pkilint.etsi.asn1 import en_319_412_1
from pkilint.itu import x520_name
from pkilint.pkix import Rfc2119Word, certificate


def _get_semantics_info_nodes(
    cert: certificate.RFC5280Certificate,
) -> List[document.PDUNode]:
    ext_and_idx = cert.get_extension_by_oid(rfc3739.id_pe_qcStatements)

    if ext_and_idx is None:
        return []

    ext, _ = ext_and_idx

    try:
        _, ext_decoded_value_node = ext.children["extnValue"].child
    except ValueError:
        return []

    nodes = []
    for statement_node in ext_decoded_value_node.children.values():
        if statement_node.children["statementId"].pdu == rfc3739.id_qcs_pkixQCSyntax_v2:
            info_node = statement_node.children.get("statementInfo")

            if info_node is not None:
                info_node_decoded = info_node.children.get("semanticsInformation")

                if info_node_decoded is not None:
                    nodes.append(info_node_decoded)

    return nodes


def _cert_has_semantics_id(
    semantics_id: univ.ObjectIdentifier, cert: certificate.RFC5280Certificate
) -> bool:
    # noinspection PyTypeChecker
    semantics_info_nodes = _get_semantics_info_nodes(cert)

    for semantics_info_node in semantics_info_nodes:
        id_oid_node = semantics_info_node.children.get("semanticsIdentifier")

        if id_oid_node is not None and id_oid_node.pdu == semantics_id:
            return True

    return False


class LegalPersonOrganizationIdentifierValidator(
    organization_id.OrganizationIdentifierValidatorBase
):
    VALIDATION_INVALID_ORGANIZATION_ID_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.4-02.invalid_format",
    )

    VALIDATION_INVALID_ORGANIZATION_ID_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.4-03.invalid_scheme",
    )

    VALIDATION_INVALID_ORGANIZATION_ID_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.4-03.invalid_country",
    )

    VALIDATION_ORGANIZATION_NATIONAL_SCHEME_DETECTED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "etsi.en_319_412_1.leg-5.1.4-03.national_identifier_scheme_detected",
    )

    _STATE_PROVINCE_PROHIBITED = (
        Rfc2119Word.MUST_NOT,
        VALIDATION_INVALID_ORGANIZATION_ID_FORMAT.code,
    )
    _REFERENCE_REQUIRED = (
        Rfc2119Word.MUST,
        VALIDATION_INVALID_ORGANIZATION_ID_FORMAT.code,
    )

    _NTR_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES,
            VALIDATION_INVALID_ORGANIZATION_ID_COUNTRY,
        ),
        state_province=_STATE_PROVINCE_PROHIBITED,
        reference=_REFERENCE_REQUIRED,
    )

    _VAT_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES
            | {
                organization_id.COUNTRY_CODE_GREECE_TRADITIONAL,
                organization_id.COUNTRY_CODE_NORTHERN_IRELAND,
            },
            VALIDATION_INVALID_ORGANIZATION_ID_COUNTRY,
        ),
        state_province=_STATE_PROVINCE_PROHIBITED,
        reference=_REFERENCE_REQUIRED,
    )

    _PSD_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES,
            VALIDATION_INVALID_ORGANIZATION_ID_COUNTRY,
        ),
        state_province=_STATE_PROVINCE_PROHIBITED,
        reference=_REFERENCE_REQUIRED,
    )

    _LEI_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            {organization_id.COUNTRY_CODE_GLOBAL_SCHEME},
            VALIDATION_INVALID_ORGANIZATION_ID_COUNTRY,
        ),
        state_province=_STATE_PROVINCE_PROHIBITED,
        reference=_REFERENCE_REQUIRED,
    )

    _ELEMENT_ALLOWANCES = {
        "NTR": _NTR_SCHEME,
        "VAT": _VAT_SCHEME,
        "PSD": _PSD_SCHEME,
        "LEI": _LEI_SCHEME,
    }

    def __init__(self):
        super().__init__(
            element_allowances=self._ELEMENT_ALLOWANCES,
            invalid_format_validation=self.VALIDATION_INVALID_ORGANIZATION_ID_FORMAT,
            additional_validations=[
                self.VALIDATION_INVALID_ORGANIZATION_ID_SCHEME,
                self.VALIDATION_ORGANIZATION_NATIONAL_SCHEME_DETECTED,
            ],
            pdu_class=x520_name.X520OrganizationIdentifier,
        )

    def match(self, node):
        if not super().match(node):
            return False

        # skip nodes that haven't been decoded
        if not any(node.children):
            return False

        # noinspection PyTypeChecker
        return _cert_has_semantics_id(
            en_319_412_1.id_etsi_qcs_SemanticsId_Legal, node.document
        )

    @classmethod
    def handle_unknown_scheme(
        cls,
        node: document.PDUNode,
        parsed: organization_id.ParsedOrganizationIdentifier,
    ):
        is_valid_national_scheme = (
            parsed.is_national_scheme
            and parsed.country
            in organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES
            and parsed.state_province is None
            and parsed.reference
        )

        value_str = str(node.child[1].pdu)

        if is_valid_national_scheme:
            raise validation.ValidationFindingEncountered(
                cls.VALIDATION_ORGANIZATION_NATIONAL_SCHEME_DETECTED,
                f'National registration scheme "{parsed.scheme}" in organization identifier: "{value_str}"',
            )
        else:
            raise validation.ValidationFindingEncountered(
                cls.VALIDATION_INVALID_ORGANIZATION_ID_SCHEME,
                f'Invalid registration scheme "{parsed.scheme}" in organization identifier: "{value_str}"',
            )

    @classmethod
    def parse_organization_id_node(
        cls, node: document.PDUNode
    ) -> ParsedOrganizationIdentifier:
        value = str(node.child[1].pdu)

        return organization_id.parse_organization_identifier(value)


_NATURAL_PERSON_ID_REGEX = re.compile(
    r"^(?P<scheme>[A-Z]{2}[A-Z:])(?P<country>[a-zA-Z]{2})-(?P<reference>.+)$"
)


class ParsedNaturalPersonIdentifier(typing.NamedTuple):
    raw: typing.Optional[str]
    scheme: str
    is_national_scheme: bool
    country: str
    reference: str


def parse_natural_person_identifier(value: str) -> ParsedNaturalPersonIdentifier:
    m = _NATURAL_PERSON_ID_REGEX.match(value)

    if m is None:
        raise ValueError(f'Invalid natural person identifier syntax: "{value}"')

    is_national_scheme = m["scheme"].endswith(":")

    scheme = m["scheme"][:2] if is_national_scheme else m["scheme"]

    return ParsedNaturalPersonIdentifier(
        value, scheme, is_national_scheme, m["country"], m["reference"]
    )


class NaturalPersonIdentifierValidator(validation.Validator):
    VALIDATION_INVALID_NATURAL_PERSON_ID_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.nat-5.1.3.invalid_identifier_format",
    )

    VALIDATION_NATURAL_PERSON_ID_NATIONAL_SCHEME_DETECTED = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.NOTICE,
            "etsi.en_319_412_1.nat-5.1.3.national_identifier_scheme_detected",
        )
    )

    VALIDATION_INVALID_NATURAL_PERSON_ID_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.nat-5.1.3.invalid_identifier_scheme",
    )

    VALIDATION_DEPRECATED_NATURAL_PERSON_ID_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "etsi.en_319_412_3.nat-5.1.3.deprecated_identifier_scheme",
    )

    VALIDATION_INVALID_NATURAL_PERSON_ID_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_419_412_4.nat-5.1.3.invalid_identifier_country",
    )

    _KNOWN_SCHEMES = {
        "PAS",
        "IDC",
        "PNO",
        "TAX",
        "TIN",
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_INVALID_NATURAL_PERSON_ID_FORMAT,
                self.VALIDATION_NATURAL_PERSON_ID_NATIONAL_SCHEME_DETECTED,
                self.VALIDATION_INVALID_NATURAL_PERSON_ID_SCHEME,
                self.VALIDATION_DEPRECATED_NATURAL_PERSON_ID_SCHEME,
                self.VALIDATION_INVALID_NATURAL_PERSON_ID_COUNTRY,
            ],
            pdu_class=rfc5280.X520SerialNumber,
        )

    def match(self, node):
        if not super().match(node):
            return False

        # noinspection PyTypeChecker
        return _cert_has_semantics_id(
            en_319_412_1.id_etsi_qcs_semanticsId_Natural, node.document
        )

    def validate(self, node):
        value = str(node.pdu)

        try:
            parsed = parse_natural_person_identifier(value)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_NATURAL_PERSON_ID_FORMAT, str(e)
            )

        findings = []
        if parsed.scheme in self._KNOWN_SCHEMES:
            if parsed.scheme == "TAX":
                findings.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_DEPRECATED_NATURAL_PERSON_ID_SCHEME,
                        f'Deprecated natural person identifier scheme: "{parsed.scheme}"',
                    )
                )
        else:
            if parsed.is_national_scheme:
                findings.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_NATURAL_PERSON_ID_NATIONAL_SCHEME_DETECTED,
                        f'National registration scheme "{parsed.scheme}" in natural person identifier: "{value}"',
                    )
                )
            else:
                findings.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_INVALID_NATURAL_PERSON_ID_SCHEME,
                        f'Invalid natural person identifier scheme: "{parsed.scheme}"',
                    )
                )

        if (
            parsed.country
            not in organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES
        ):
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_INVALID_NATURAL_PERSON_ID_COUNTRY,
                    f'Invalid natural person identifier country code: "{parsed.country}"',
                )
            )

        return validation.ValidationResult(self, node, findings)


class EidasLegalPersonIdentifierValidator(validation.Validator):
    """
    LEG-5.1.6-03: Any organizationIdentifier attribute present in the subject field of the certificate shall
    comply with the content requirement specified for the eIDAS LegalPersonIdentifier attribute.

    From eIDAS SAML Attribute Profile v1.2 Final, section 2.5:
    - The Unique Identifier MUST NOT contain any whitespace.
    - The Unique Identifier MUST NOT exceed a total of 256 characters.
    """

    VALIDATION_EIDAS_LEGAL_PERSON_IDENTIFIER_WHITESPACE_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.6-03.eidas_legal_person_identifier_whitespace_present",
    )

    VALIDATION_EIDAS_LEGAL_PERSON_IDENTIFIER_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.6-03.eidas_legal_person_identifier_too_long",
    )

    _MAX_LENGTH = 256

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_EIDAS_LEGAL_PERSON_IDENTIFIER_WHITESPACE_PRESENT,
                self.VALIDATION_EIDAS_LEGAL_PERSON_IDENTIFIER_TOO_LONG,
            ],
            pdu_class=x520_name.X520OrganizationIdentifier,
        )

    def match(self, node):
        # noinspection PyTypeChecker
        return super().match(node) and _cert_has_semantics_id(
            en_319_412_1.id_etsi_qcs_SemanticsId_eIDASLegal, node.document
        )

    def validate(self, node):
        value = str(node.child[1].pdu)

        if any(c.isspace() for c in value):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EIDAS_LEGAL_PERSON_IDENTIFIER_WHITESPACE_PRESENT,
                f'Whitespace present in organization identifier: "{value}"',
            )

        value_len = len(value)

        if value_len > self._MAX_LENGTH:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EIDAS_LEGAL_PERSON_IDENTIFIER_TOO_LONG,
                f'Organization identifier "{value}" ({value_len} characters) exceeds maximum length of '
                f"{self._MAX_LENGTH} characters",
            )


class NameRegistrationAuthoritiesValidatorBase(validation.Validator):
    def __init__(
        self,
        name_registration_authorities_missing_validation: validation.ValidationFinding,
        name_registration_authorities_uri_missing_validation: validation.ValidationFinding,
        attribute_type_id: univ.ObjectIdentifier,
        semantics_id: univ.ObjectIdentifier,
    ):
        self._name_registration_authorities_missing_validation = (
            name_registration_authorities_missing_validation
        )
        self._name_registration_authorities_uri_missing_validation = (
            name_registration_authorities_uri_missing_validation
        )

        self._attribute_type_id = attribute_type_id
        self._semantics_id = semantics_id

        validations = [
            name_registration_authorities_missing_validation,
            name_registration_authorities_uri_missing_validation,
        ]

        super().__init__(
            validations=validations, pdu_class=rfc3739.SemanticsInformation
        )

    @classmethod
    def is_attribute_value_national_id_scheme(cls, atv_node: document.PDUNode) -> bool:
        pass

    def match(self, node):
        if not super().match(node):
            return False

        semantics_id_node = node.children.get("semanticsIdentifier")

        if semantics_id_node is None or semantics_id_node.pdu != self._semantics_id:
            return False

        atvs_and_idxs = node.document.get_subject_attributes_by_type(
            self._attribute_type_id
        )

        return any(
            self.is_attribute_value_national_id_scheme(a) for a, _ in atvs_and_idxs
        )

    def validate(self, node):
        nra_node = node.children.get("nameRegistrationAuthorities")

        if nra_node is None:
            raise validation.ValidationFindingEncountered(
                self._name_registration_authorities_missing_validation
            )

        if not any(
            g.child[0] == "uniformResourceIdentifier"
            for g in nra_node.children.values()
        ):
            raise validation.ValidationFindingEncountered(
                self._name_registration_authorities_uri_missing_validation
            )


class NaturalPersonIdentifierNameRegistrationAuthoritiesValidator(
    NameRegistrationAuthoritiesValidatorBase
):
    """
    NAT-5.1.3-05: When a locally defined identity type reference is provided (two characters followed by ":"), the
    nameRegistrationAuthorities element of SemanticsInformation (IETF RFC 3739 [1]) shall be present.
    """

    VALIDATION_NAME_REGISTRATION_AUTHORITIES_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.nat-5.1.3-05.national_id_scheme_name_registration_authorities_missing",
    )

    """
    NAT-5.1.3-06: The nameRegistrationAuthorities element of SemanticsInformation (IETF RFC 3739 [1]) shall
    contain at least a uniformResourceIdentifier generalName.
    """
    VALIDATION_NAME_REGISTRATION_AUTHORITIES_URI_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.nat-5.1.3-05.national_id_scheme_name_registration_authorities_uri_missing",
    )

    def __init__(self):
        super().__init__(
            self.VALIDATION_NAME_REGISTRATION_AUTHORITIES_MISSING,
            self.VALIDATION_NAME_REGISTRATION_AUTHORITIES_URI_MISSING,
            rfc5280.id_at_serialNumber,
            en_319_412_1.id_etsi_qcs_semanticsId_Natural,
        )

    @classmethod
    def is_attribute_value_national_id_scheme(cls, atv_node: document.PDUNode) -> bool:
        value_node = atv_node.children["value"]

        try:
            _, decoded_value_node = value_node.child

            value_str = str(decoded_value_node.pdu)

            parsed = parse_natural_person_identifier(value_str)
        except ValueError:
            return False

        return parsed.is_national_scheme


class LegalPersonIdentifierNameRegistrationAuthoritiesValidator(
    NameRegistrationAuthoritiesValidatorBase
):
    """
    LEG-5.1.4-05: When a locally defined identity type reference is provided (two characters followed by ":"), the
    nameRegistrationAuthorities element of SemanticsInformation (IETF RFC 3739 [1]) shall be present and
    shall contain at least a uniformResourceIdentifier generalName.
    """

    VALIDATION_NAME_REGISTRATION_AUTHORITIES_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.4-05.national_id_scheme_name_registration_authorities_missing",
    )

    VALIDATION_NAME_REGISTRATION_AUTHORITIES_URI_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_1.leg-5.1.4-05.national_id_scheme_name_registration_authorities_uri_missing",
    )

    def __init__(self):
        super().__init__(
            self.VALIDATION_NAME_REGISTRATION_AUTHORITIES_MISSING,
            self.VALIDATION_NAME_REGISTRATION_AUTHORITIES_URI_MISSING,
            x520_name.id_at_organizationIdentifier,
            en_319_412_1.id_etsi_qcs_SemanticsId_Legal,
        )

    @classmethod
    def is_attribute_value_national_id_scheme(cls, atv_node: document.PDUNode) -> bool:
        value_node = atv_node.children["value"]

        try:
            _, decoded_dirstring_node = value_node.child

            _, decoded_value_node = decoded_dirstring_node.child

            value_str = str(decoded_value_node.pdu)

            parsed = organization_id.parse_organization_identifier(value_str)
        except ValueError:
            return False

        return parsed.is_national_scheme


class NaturalPersonEidasIdentifierValidator(validation.Validator):
    """
    NAT-5.1.5-01: If using electronic identity attributes as specified in eIDAS SAML
    attribute profile for a certificate issued to natural persons, the semantics of
    id-etsi-qcs-SemanticsId-eIDASNatural shall be as follows.
    NAT-5.1.5-02: If the eIDAS natural person identifier is included, the values
    of attributes in the subject field shall meet the content requirements of corresponding
    attributes defined by the eIDAS SAML attribute profile according to the following requirements.

    NAT-5.1.5-03: Any serialNumber attribute value in the subject field of the certificate
    shall comply with the content requirement specified for the eIDAS PersonIdentifier attribute.

    NAT-5.1.5-04: Attributes present in subject field of the certificate are equivalent to defined attributes
    in accordance to the below table. This means that the present attribute shall hold equivalent
    information, even if the format used to express that information differs.

    serialNumber -> PersonIdentifier Example: ES/AT/02635542Y
    """

    VALIDATION_INVALID_ISO_3166 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "etsi.nat_5.1.5-03.invalid_iso_3166"
    )

    VALIDATION_INVALID_SYNTAX_SERIAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.nat_5.1.5-03.invalid_syntax_serial",
    )
    VALIDATION_INVALID_CHARACTER_SERIAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.nat_5.1.5-03.invalid_character_serial",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_INVALID_ISO_3166,
                self.VALIDATION_INVALID_SYNTAX_SERIAL,
                self.VALIDATION_INVALID_CHARACTER_SERIAL,
            ],
            pdu_class=rfc5280.X520SerialNumber,
        )

    def match(self, node):
        if not super().match(node):
            return False

        # noinspection PyTypeChecker
        return _cert_has_semantics_id(
            en_319_412_1.id_etsi_qcs_semanticsId_eIDASNatural, node.document
        )

    def validate(self, node):
        value = str(node.pdu)
        findings = []
        if len(value) < 7:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_SYNTAX_SERIAL,
                "Invalid serial number syntax (needs more characters).",
            )
        if value[2] != "/" or value[5] != "/":
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_INVALID_CHARACTER_SERIAL,
                    "No backslash found after ISO-3166 country code.",
                )
            )
        if (
            value[0:2] not in organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES
            or value[3:5]
            not in organization_id.ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES
        ):
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_INVALID_ISO_3166, "Invalid 3166 ISO Code."
                )
            )

        return validation.ValidationResult(self, node, findings)



#!/usr/bin/env python

import argparse
import sys

from pkilint import pkix
from pkilint import cli_util, loader, report
from pkilint.pkix import extension, name, ocsp


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(description="RFC 6960 OCSP Response Linter")

    subparsers = parser.add_subparsers(dest="command", required=True)
    subparsers.add_parser(
        "validations", help="Output the set of validations which this linter performs"
    )

    lint_parser = subparsers.add_parser("lint", help="Lint the specified OCSP response")
    cli_util.add_standard_args(lint_parser)

    lint_parser.add_argument(
        "file", type=argparse.FileType("rb"), help="The OCSP response to lint"
    )

    args = parser.parse_args(cli_args)

    doc_validator = ocsp.create_pkix_ocsp_response_validator_container(
        [
            ocsp.create_response_decoder(),
            pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
            pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        ],
        [],
    )

    if args.command == "validations":
        print(report.report_included_validations(doc_validator))

        return 0
    else:
        try:
            ocsp_response = (
                loader.RFC6960OCSPResponseDocumentLoader().get_file_loader_func(
                    args.document_format
                )(args.file, args.file.name)
            )
        except ValueError as e:
            print(f"Failed to load OCSP response: {e}", file=sys.stderr)
            return 1

        results = doc_validator.validate(ocsp_response.root)

        print(args.format(results, args.severity))

        return cli_util.clamp_exit_code(
            report.get_findings_count(results, args.severity)
        )


if __name__ == "__main__":
    sys.exit(main())



import collections
from typing import List, Set

import validators
from pyasn1.type.univ import ObjectIdentifier
from pyasn1_alt_modules import rfc5280, rfc2985

from pkilint import document, validation, oid
from pkilint.itu import x520_name

ATTRIBUTE_TYPE_MAPPINGS = {
    **x520_name.ATTRIBUTE_TYPE_MAPPINGS,
    **rfc2985._certificateAttributesMapUpdate,
    **rfc5280.certificateAttributesMap,
}


def get_name_attributes_by_type(name_node, type_oid):
    atvs = []

    for rdn_idx, rdn in name_node.children["rdnSequence"].children.items():
        for atv_idx, atv in rdn.children.items():
            if atv.children["type"].pdu == type_oid:
                atvs.append((atv, (int(rdn_idx), int(atv_idx))))

    return atvs


def get_name_attribute_counts(name_node):
    counts = collections.Counter()

    for rdn_idx, rdn in name_node.children["rdnSequence"].children.items():
        counts.update((atv.children["type"].pdu for atv in rdn.children.values()))

    return counts


class EmptyNameValidator(validation.Validator):
    VALIDATION_NAME_IS_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.name_empty"
    )

    def __init__(self, **kwargs):
        super().__init__(
            validations=[self.VALIDATION_NAME_IS_EMPTY],
            pdu_class=rfc5280.RDNSequence,
            **kwargs,
        )

    def validate(self, node):
        if len(node.children) == 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NAME_IS_EMPTY,
            )


class RDNContainsUniqueTypesValidator(validation.Validator):
    VALIDATION_ATTRIBUTE_TYPES_NOT_UNIQUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.rdn_contains_duplicate_attribute_types",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_ATTRIBUTE_TYPES_NOT_UNIQUE,
            pdu_class=rfc5280.RelativeDistinguishedName,
        )

    def validate(self, node):
        oids = set()
        for child in node.children.values():
            oid = child.children["type"].pdu
            if oid in oids:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_ATTRIBUTE_TYPES_NOT_UNIQUE,
                    f'Multiple attributes of type "{str(oid)}"',
                )

            oids.add(oid)


class NameAttributeTypesValidatorBase(validation.Validator):
    def __init__(
        self,
        *,
        expected_oid_set: Set[ObjectIdentifier],
        validation: validation.ValidationFinding,
        **kwargs,
    ):
        self.expected_oid_set = expected_oid_set
        self.validation = validation

        super().__init__(
            validations=[self.validation], pdu_class=rfc5280.RDNSequence, **kwargs
        )

    def validate_attributes(
        self, node: document.PDUNode, attributes: List[ObjectIdentifier]
    ) -> validation.ValidationResult:
        pass

    def validate(self, node):
        attributes = []

        for rdn in node.children.values():
            attributes += [atv.children["type"].pdu for atv in rdn.children.values()]

        return self.validate_attributes(node, attributes)


class PermittedAttributeTypeValidator(NameAttributeTypesValidatorBase):
    def __init__(self, *, allowed_oid_set, validation):
        super().__init__(expected_oid_set=allowed_oid_set, validation=validation)

    def validate_attributes(self, node, attributes):
        attributes = set(attributes)

        prohibited_oids = attributes - self.expected_oid_set

        if len(prohibited_oids) > 0:
            oids_str = oid.format_oids(prohibited_oids)

            raise validation.ValidationFindingEncountered(
                self.validation, f"Prohibited attribute types: {oids_str}"
            )


class RequiredAttributeTypeValidator(NameAttributeTypesValidatorBase):
    def __init__(self, *, required_oid_set, validation):
        super().__init__(expected_oid_set=required_oid_set, validation=validation)

    def validate_attributes(self, node, attributes):
        attributes = set(attributes)

        missing_oids = self.expected_oid_set - attributes

        if len(missing_oids) > 0:
            oids_str = oid.format_oids(missing_oids)

            raise validation.ValidationFindingEncountered(
                self.validation, f"Required attribute types not present: {oids_str}"
            )


class NameDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(
            decode_func=decode_func, pdu_class=rfc5280.AttributeTypeAndValue, **kwargs
        )


class IssuerSubjectNameBinaryEqualValidator(validation.DEREqualityValidator):
    VALIDATION_ISSUER_AND_SUBJECT_NOT_DER_EQUAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "pkix.issuer_and_subject_dn_not_binary_equal",
    )

    def _retrieve_subject_issuer_dn(self, node):
        return node.navigate(self._subject_document_issuer_dn_path)

    def __init__(self, *, subject_document_issuer_dn_path, **kwargs):
        self._subject_document_issuer_dn_path = subject_document_issuer_dn_path

        super().__init__(
            other_node_retriever=self._retrieve_subject_issuer_dn,
            validation=self.VALIDATION_ISSUER_AND_SUBJECT_NOT_DER_EQUAL,
            **kwargs,
        )


class DomainComponentValidDomainNameValidator(validation.Validator):
    VALIDATION_NAME_DC_NOT_A_VALID_DOMAIN_NAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.name_domain_components_invalid_domain_name",
    )

    def __init__(self, **kwargs):
        super().__init__(
            validations=[self.VALIDATION_NAME_DC_NOT_A_VALID_DOMAIN_NAME], **kwargs
        )

    def validate(self, node):
        atvs = node.document.get_name_attributes_by_type(
            rfc5280.id_domainComponent, f":{node.path}"
        )

        try:
            components = [str(a.navigate("value.domainComponent").pdu) for a, _ in atvs]
        except document.PDUNavigationFailedError:
            return

        components.reverse()

        domain_name = ".".join(components)

        if len(domain_name) > 0:
            ret = validators.domain(domain_name)

            if not isinstance(ret, bool) or not ret:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_NAME_DC_NOT_A_VALID_DOMAIN_NAME,
                    f'Invalid domain name in domainComponents: "{domain_name}"',
                )


class DuplicateAttributeTypeValidator(NameAttributeTypesValidatorBase):
    def __init__(self, *, allowed_duplicate_oid_set, validation):
        super().__init__(
            expected_oid_set=allowed_duplicate_oid_set,
            validation=validation,
        )

    def validate_attributes(self, node, attributes):
        duplicates = set((o for o in attributes if attributes.count(o) > 1))

        disallowed_duplicates = duplicates - self.expected_oid_set

        if len(disallowed_duplicates) > 0:
            oids_str = oid.format_oids(disallowed_duplicates)

            raise validation.ValidationFindingEncountered(
                self.validation,
                f"Prohibited multiple instances of attribute types: {oids_str}",
            )



import re

import validators
from pyasn1_alt_modules import rfc5280, rfc8398

from pkilint import validation, pkix, oid
from pkilint.cabf import cabf_name
from pkilint.cabf.cabf_name import CabfOrganizationIdentifierAttributeValidator
from pkilint.cabf.smime.smime_constants import Generation, ValidationLevel
from pkilint.common import organization_id
from pkilint.common.organization_id import OrganizationIdentifierLeiValidator
from pkilint.itu import x520_name, asn1_util
from pkilint.pkix import certificate, name, Rfc2119Word, general_name

SHALL = pkix.Rfc2119Word.SHALL
SHALL_NOT = pkix.Rfc2119Word.SHALL_NOT
MAY = pkix.Rfc2119Word.MAY

_MV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_organizationalUnitName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_organizationIdentifier: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_givenName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_surname: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_pseudonym: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_streetAddress: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_localityName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_stateOrProvinceName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_postalCode: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_countryName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
}

_OV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL, SHALL, SHALL),
    rfc5280.id_at_organizationalUnitName: (MAY, MAY, MAY),
    x520_name.id_at_organizationIdentifier: (SHALL, SHALL, SHALL),
    rfc5280.id_at_givenName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_surname: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_pseudonym: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_streetAddress: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_localityName: (MAY, MAY, MAY),
    rfc5280.id_at_stateOrProvinceName: (MAY, MAY, MAY),
    x520_name.id_at_postalCode: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_countryName: (MAY, MAY, MAY),
}

_SV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL, SHALL, SHALL),
    rfc5280.id_at_organizationalUnitName: (MAY, MAY, MAY),
    x520_name.id_at_organizationIdentifier: (SHALL, SHALL, SHALL),
    rfc5280.id_at_givenName: (MAY, MAY, MAY),
    rfc5280.id_at_surname: (MAY, MAY, MAY),
    rfc5280.id_at_pseudonym: (MAY, MAY, MAY),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (MAY, MAY, MAY),
    x520_name.id_at_streetAddress: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_localityName: (MAY, MAY, MAY),
    rfc5280.id_at_stateOrProvinceName: (MAY, MAY, MAY),
    x520_name.id_at_postalCode: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_countryName: (MAY, MAY, MAY),
}

_IV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_organizationalUnitName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_organizationIdentifier: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_givenName: (MAY, MAY, MAY),
    rfc5280.id_at_surname: (MAY, MAY, MAY),
    rfc5280.id_at_pseudonym: (MAY, MAY, MAY),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (MAY, MAY, MAY),
    x520_name.id_at_streetAddress: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_localityName: (MAY, MAY, MAY),
    rfc5280.id_at_stateOrProvinceName: (MAY, MAY, MAY),
    x520_name.id_at_postalCode: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_countryName: (MAY, MAY, MAY),
}

_GENERATION_INDEXES = {
    Generation.LEGACY: 0,
    Generation.MULTIPURPOSE: 1,
    Generation.STRICT: 2,
}

_VALIDATION_LEVEL_TO_TABLE = {
    ValidationLevel.MAILBOX: _MV_ATTRIBUTES,
    ValidationLevel.ORGANIZATION: _OV_ATTRIBUTES,
    ValidationLevel.SPONSORED: _SV_ATTRIBUTES,
    ValidationLevel.INDIVIDUAL: _IV_ATTRIBUTES,
}

_VALIDATION_LEVEL_TO_OTHER_ATTRIBUTE_ALLOWANCE = {
    ValidationLevel.MAILBOX: (False, False, False),
    ValidationLevel.ORGANIZATION: (True, False, False),
    ValidationLevel.SPONSORED: (True, False, False),
    ValidationLevel.INDIVIDUAL: (True, False, False),
}

_REQUIRED_ONE_OF_N = {
    (ValidationLevel.SPONSORED, Generation.LEGACY): {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
        rfc5280.id_at_commonName,
    },
    (ValidationLevel.INDIVIDUAL, Generation.LEGACY): {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
        rfc5280.id_at_commonName,
    },
    (ValidationLevel.SPONSORED, Generation.MULTIPURPOSE): {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
    },
    (ValidationLevel.INDIVIDUAL, Generation.MULTIPURPOSE): {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
    },
    (ValidationLevel.SPONSORED, Generation.STRICT): {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
    },
    (ValidationLevel.INDIVIDUAL, Generation.STRICT): {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
    },
}


class SubscriberSubjectValidator(validation.Validator):
    VALIDATION_MISSING_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.missing_required_attribute",
    )

    VALIDATION_PROHIBITED_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "cabf.smime.prohibited_attribute"
    )

    VALIDATION_MIXED_NAME_AND_PSEUDONYM_ATTRIBUTES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.mixed_name_and_pseudonym_attributes",
    )

    def __init__(self, validation_level, generation):
        super().__init__(
            validations=[
                self.VALIDATION_PROHIBITED_ATTRIBUTE,
                self.VALIDATION_MISSING_ATTRIBUTE,
                self.VALIDATION_MIXED_NAME_AND_PSEUDONYM_ATTRIBUTES,
            ],
            pdu_class=rfc5280.RDNSequence,
            predicate=lambda n: n.path
            != "certificate.tbsCertificate.issuer.rdnSequence",
        )

        self._attribute_table = {
            k: v[_GENERATION_INDEXES[generation]]
            for k, v in _VALIDATION_LEVEL_TO_TABLE[validation_level].items()
        }

        self._required_attributes = {
            k for k, v in self._attribute_table.items() if v == SHALL
        }
        self._prohibited_attributes = {
            k for k, v, in self._attribute_table.items() if v == SHALL_NOT
        }
        self._required_one_of_n_attributes = _REQUIRED_ONE_OF_N.get(
            (validation_level, generation)
        )

        self._allow_other_oids = _VALIDATION_LEVEL_TO_OTHER_ATTRIBUTE_ALLOWANCE[
            validation_level
        ][_GENERATION_INDEXES[generation]]

    def validate(self, node):
        findings = []

        attributes = set()
        for rdn in node.children.values():
            attributes.update(
                (atv.children["type"].pdu for atv in rdn.children.values())
            )

        findings.extend(
            (
                validation.ValidationFindingDescription(
                    self.VALIDATION_MISSING_ATTRIBUTE,
                    f"Missing required attribute: {a}",
                )
                for a in self._required_attributes - attributes
            )
        )

        if (
            self._required_one_of_n_attributes
            and len(self._required_one_of_n_attributes.intersection(attributes)) == 0
        ):
            oids = oid.format_oids(self._required_one_of_n_attributes)

            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_MISSING_ATTRIBUTE,
                    f"Missing one of these required attributes: {oids}",
                )
            )

        findings.extend(
            (
                validation.ValidationFindingDescription(
                    self.VALIDATION_PROHIBITED_ATTRIBUTE, f"Prohibited attribute: {a}"
                )
                for a in self._prohibited_attributes.intersection(attributes)
            )
        )

        if rfc5280.id_at_pseudonym in attributes and (
            any(
                {rfc5280.id_at_givenName, rfc5280.id_at_surname}.intersection(
                    attributes
                )
            )
        ):
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_MIXED_NAME_AND_PSEUDONYM_ATTRIBUTES, None
                )
            )

        if not self._allow_other_oids:
            findings.extend(
                (
                    validation.ValidationFindingDescription(
                        self.VALIDATION_PROHIBITED_ATTRIBUTE,
                        f"Prohibited other attribute: {a}",
                    )
                    for a in attributes - set(self._attribute_table.keys())
                )
            )

        return validation.ValidationResult(self, node, findings)


class CabfSmimeOrganizationIdentifierAttributeValidator(
    CabfOrganizationIdentifierAttributeValidator
):
    _REFERENCE_PROHIBITED = (
        Rfc2119Word.MUST_NOT,
        "cabf.smime.prohibited_organization_identifier_reference_present_for_scheme",
    )

    """
    From SMBR 7.1.4.2.2 (d):
   
    When the Organization or Legal Entity is registered in Germany, the Registration Reference SHOULD use the EUID
    identifier.
    """
    VALIDATION_GERMAN_NTR_REFERENCE_NOT_EUID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.smime.german_ntr_registration_reference_not_euid",
    )

    _LEI_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            {organization_id.COUNTRY_CODE_GLOBAL_SCHEME},
            CabfOrganizationIdentifierAttributeValidator.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
        ),
        state_province=CabfOrganizationIdentifierAttributeValidator.STATE_PROVINCE_PROHIBITED,
        reference=CabfOrganizationIdentifierAttributeValidator.REFERENCE_REQUIRED,
    )
    _GOV_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            organization_id.ISO3166_1_COUNTRY_CODES,
            CabfOrganizationIdentifierAttributeValidator.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
        ),
        state_province=(Rfc2119Word.MAY, None),
        reference=_REFERENCE_PROHIBITED,
    )
    _INT_SCHEME = organization_id.OrganizationIdentifierElementAllowance(
        country_codes=(
            {organization_id.COUNTRY_CODE_GLOBAL_SCHEME},
            CabfOrganizationIdentifierAttributeValidator.VALIDATION_ORGANIZATION_ID_INVALID_COUNTRY,
        ),
        state_province=CabfOrganizationIdentifierAttributeValidator.STATE_PROVINCE_PROHIBITED,
        reference=_REFERENCE_PROHIBITED,
    )

    _EUID_SYNTAX = re.compile(
        r"^(?P<country>[A-Z]{2})(?P<register>.+)\.(?P<reference>.+)$"
    )

    def __init__(self):
        super().__init__(
            {
                "LEI": self._LEI_SCHEME,
                "GOV": self._GOV_SCHEME,
                "INT": self._INT_SCHEME,
            },
            enforce_strict_state_province_format=False,
            additional_validations=[self.VALIDATION_GERMAN_NTR_REFERENCE_NOT_EUID],
        )

    def validate_with_parsed_value(self, node, parsed):
        result = super().validate_with_parsed_value(node, parsed)

        if any(result.finding_descriptions):
            return validation.ValidationResult(self, node, result.finding_descriptions)

        if parsed.scheme == "NTR" and parsed.country == "DE":
            if parsed.state_province is not None:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_GERMAN_NTR_REFERENCE_NOT_EUID,
                    f'Organization identifier contains state/province: "{parsed.raw}"',
                )

            m = self._EUID_SYNTAX.match(parsed.reference)

            if m is None:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_GERMAN_NTR_REFERENCE_NOT_EUID,
                    f'Registration Reference is not in EUID format: "{parsed.reference}"',
                )

            if m["country"] != parsed.country:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_GERMAN_NTR_REFERENCE_NOT_EUID,
                    f'EUID Registration Reference has mismatched country code: "{parsed.raw}"',
                )


class SubscriberAttributeDependencyValidator(validation.Validator):
    VALIDATION_MISSING_REQUIRED_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.required_attribute_missing_for_dependent_attribute",
    )

    _ATTRIBUTE_DEPENDENCIES = [
        (
            x520_name.id_at_streetAddress,
            {rfc5280.id_at_localityName, rfc5280.id_at_stateOrProvinceName},
        ),
        (rfc5280.id_at_stateOrProvinceName, {rfc5280.id_at_countryName}),
        (rfc5280.id_at_localityName, {rfc5280.id_at_countryName}),
        (x520_name.id_at_postalCode, {rfc5280.id_at_countryName}),
    ]

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_MISSING_REQUIRED_ATTRIBUTE],
            pdu_class=rfc5280.RDNSequence,
            predicate=lambda n: n.path
            != "certificate.tbsCertificate.issuer.rdnSequence",
        )

    def validate(self, node):
        attributes = set()
        for rdn in node.children.values():
            attributes.update(
                (atv.children["type"].pdu for atv in rdn.children.values())
            )

        for dependent_attribute, required_attributes in self._ATTRIBUTE_DEPENDENCIES:
            if dependent_attribute in attributes:
                if not attributes & required_attributes:
                    oids = oid.format_oids(required_attributes)

                    if len(required_attributes) > 1:
                        message = f"one of {oids} is not present"
                    else:
                        message = f"{oids} is not present"

                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_MISSING_REQUIRED_ATTRIBUTE,
                        f"{dependent_attribute} is present but {message}",
                    )


def create_subscriber_certificate_subject_validator_container(
    validation_level, generation
):
    dn_validators = [
        SubscriberSubjectValidator(validation_level, generation),
        SubscriberAttributeDependencyValidator(),
        SubjectAlternativeNameContainsSubjectEmailAddressesValidator(),
        cabf_name.ValidCountryValidator(),
        CommonNameValidator(validation_level, generation),
        CabfSmimeOrganizationIdentifierAttributeValidator(),
        OrganizationIdentifierLeiValidator(),
        OrganizationIdentifierCountryNameConsistentValidator(),
        cabf_name.RelativeDistinguishedNameContainsOneElementValidator(),
        cabf_name.SignificantAttributeValueValidator(),
        cabf_name.HTMLEntitiesValidator(),
    ]

    return certificate.create_subject_validator_container(
        dn_validators,
        pdu_class=rfc5280.Name,
        predicate=lambda n: n.path != "certificate.tbsCertificate.issuer",
    )


class SubjectAlternativeNameContainsSubjectEmailAddressesValidator(
    validation.Validator
):
    VALIDATION_EMAIL_ADDRESS_IN_ATTRIBUTE_MISSING_FROM_SAN = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "cabf.smime.email_address_in_attribute_not_in_san",
        )
    )

    VALIDATION_UNPARSED_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "cabf.smime.unparsed_attribute_value_encountered",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_EMAIL_ADDRESS_IN_ATTRIBUTE_MISSING_FROM_SAN,
                self.VALIDATION_UNPARSED_ATTRIBUTE,
            ],
            pdu_class=rfc5280.AttributeTypeAndValue,
            # emailAddress presence in SAN is checked by PKIX lint
            predicate=lambda n: n.children["type"].pdu != rfc5280.id_emailAddress,
        )

    def validate(self, node):
        oid = node.children["type"].pdu

        value_str = asn1_util.get_string_value_from_attribute_node(node)

        if value_str is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNPARSED_ATTRIBUTE,
                f"Unparsed attribute {str(oid)} encountered",
            )

        if bool(validators.email(value_str)):
            san_email_addresses = get_email_addresses_from_san(node.document)

            if value_str not in san_email_addresses:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_EMAIL_ADDRESS_IN_ATTRIBUTE_MISSING_FROM_SAN,
                    f'Attribute {str(oid)} with value "{value_str}" not found in SAN',
                )


class CommonNameValidator(validation.Validator):
    VALIDATION_COMMON_NAME_UNKNOWN_VALUE_SOURCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.common_name_value_unknown_source",
    )

    VALIDATION_UNPARSED_COMMON_NAME_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "cabf.smime.unparsed_common_name_value",
    )

    def __init__(self, validation_level, generation):
        super().__init__(
            validations=[
                self.VALIDATION_COMMON_NAME_UNKNOWN_VALUE_SOURCE,
                self.VALIDATION_UNPARSED_COMMON_NAME_VALUE,
            ],
            pdu_class=rfc5280.X520CommonName,
        )

        self._validation_level = validation_level
        self._generation = generation

    @staticmethod
    def _is_value_in_dirstring_atvs(atvs, expected_value_node):
        expected_value_str = str(expected_value_node.pdu)

        return any(
            expected_value_str == asn1_util.get_string_value_from_attribute_node(a)
            for a in atvs
        )

    def validate(self, node):
        try:
            _, cn_value_node = node.child
        except ValueError:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNPARSED_COMMON_NAME_VALUE
            )

        parent_name_node = next(
            (n for n in node.parents if isinstance(n.pdu, rfc5280.Name))
        )

        if self._validation_level in {
            ValidationLevel.SPONSORED,
            ValidationLevel.INDIVIDUAL,
        }:
            # we don't need the index
            pseudonym_nodes = [
                t[0]
                for t in name.get_name_attributes_by_type(
                    parent_name_node, rfc5280.id_at_pseudonym
                )
            ]

            # Legacy sponsored and individual profiles allow the Personal Name in CN without being in other
            # subject attributes. However, if any pseudonym attributes are present, the CN must not contain a Personal
            # Name.
            if self._generation == Generation.LEGACY and not pseudonym_nodes:
                return

            if CommonNameValidator._is_value_in_dirstring_atvs(
                pseudonym_nodes, cn_value_node
            ):
                return

            # if there's a GN or SN, assume it's in the CN
            if any(
                name.get_name_attributes_by_type(
                    parent_name_node, rfc5280.id_at_givenName
                )
            ) or any(
                name.get_name_attributes_by_type(
                    parent_name_node, rfc5280.id_at_surname
                )
            ):
                return
        elif self._validation_level == ValidationLevel.ORGANIZATION:
            orgname_nodes = [
                t[0]
                for t in name.get_name_attributes_by_type(
                    parent_name_node, rfc5280.id_at_organizationName
                )
            ]

            if CommonNameValidator._is_value_in_dirstring_atvs(
                orgname_nodes, cn_value_node
            ):
                return

        email_addresses = get_email_addresses_from_san(node.document)

        if str(cn_value_node.pdu) not in email_addresses:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_COMMON_NAME_UNKNOWN_VALUE_SOURCE,
                f'Unknown CN value source: "{str(cn_value_node.pdu)}"',
            )


class OrganizationIdentifierCountryNameConsistentValidator(validation.Validator):
    VALIDATION_ORGID_COUNTRYNAME_INCONSISTENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.smime.org_identifier_and_country_name_attribute_inconsistent",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_ORGID_COUNTRYNAME_INCONSISTENT,
            pdu_class=rfc5280.X520countryName,
        )

    def validate(self, node):
        country_name_value = str(node.pdu)

        for atv, _ in node.document.get_subject_attributes_by_type(
            x520_name.id_at_organizationIdentifier
        ):
            x520_value_str = asn1_util.get_string_value_from_attribute_node(atv)

            if x520_value_str is None:
                continue

            try:
                parsed_org_id = organization_id.parse_organization_identifier(
                    x520_value_str
                )
            except ValueError:
                continue

            orgid_country_name = parsed_org_id.country

            # skip this orgId attribute if it contains the global scheme identifier
            if orgid_country_name == organization_id.COUNTRY_CODE_GLOBAL_SCHEME:
                continue

            if orgid_country_name != country_name_value:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_ORGID_COUNTRYNAME_INCONSISTENT,
                    f'CountryName attribute value: "{country_name_value}", '
                    f'OrganizationIdentifier attribute country name value: "{orgid_country_name}"',
                )


def get_email_addresses_from_san(cert_document):
    san_ext_and_idx = cert_document.get_extension_by_oid(rfc5280.id_ce_subjectAltName)

    if san_ext_and_idx is None:
        return []

    san_ext, _ = san_ext_and_idx

    email_addresses = []
    for gn in san_ext.navigate("extnValue.subjectAltName").children.values():
        name, value = gn.child

        if name == general_name.GeneralNameTypeName.RFC822_NAME:
            email_addresses.append(value.pdu)
        elif (
            name == general_name.GeneralNameTypeName.OTHER_NAME
            and value.navigate("type-id").pdu == rfc8398.id_on_SmtpUTF8Mailbox
        ):
            email_addresses.append(value.navigate("value").child[1].pdu)

    return email_addresses



import re
from typing import NamedTuple, Optional, Set, Tuple, Dict, List

from iso3166 import countries_by_alpha2

from pkilint import validation, document
from pkilint.iso import lei
from pkilint.itu import x520_name
from pkilint.pkix import Rfc2119Word

_ORG_ID_REGEX = re.compile(
    r"^(?P<scheme>[A-Z]{2}[A-Z:])(?P<country>[a-zA-Z]{2})(\+(?P<sp>[a-zA-Z0-9]{1,3}))?"
    r"(-(?P<reference>.+))?$"
)

# alternative country codes
COUNTRY_CODE_GREECE_TRADITIONAL = "EL"
COUNTRY_CODE_NORTHERN_IRELAND = "XI"

# multi-national codes
COUNTRY_CODE_GLOBAL_SCHEME = "XG"
COUNTRY_CODE_EUROPEAN_UNION = "EU"
COUNTRY_CODE_EUROZONE = "EZ"
COUNTRY_CODE_USSR = "SU"
COUNTRY_CODE_UNITED_NATIONS = "UN"

TRANSNATIONAL_COUNTRY_CODES = {
    COUNTRY_CODE_EUROPEAN_UNION,
    COUNTRY_CODE_EUROZONE,
    COUNTRY_CODE_USSR,
    COUNTRY_CODE_UNITED_NATIONS,
}

ISO3166_1_COUNTRY_CODES = set(countries_by_alpha2.keys())
ISO3166_1_WITH_TRANSNATIONAL_COUNTRY_CODES = (
    ISO3166_1_COUNTRY_CODES | TRANSNATIONAL_COUNTRY_CODES
)


LEI_PREFIX = "LEIXG-"


class ParsedOrganizationIdentifier(NamedTuple):
    raw: Optional[str]
    scheme: str
    is_national_scheme: bool
    country: str
    state_province: Optional[str]
    reference: Optional[str]


def parse_organization_identifier(value: str) -> ParsedOrganizationIdentifier:
    m = _ORG_ID_REGEX.match(value)

    if m is None:
        raise ValueError(f'Invalid organization identifier syntax: "{value}"')

    is_national_scheme = m["scheme"].endswith(":")

    scheme = m["scheme"][:2] if is_national_scheme else m["scheme"]

    return ParsedOrganizationIdentifier(
        value,
        scheme,
        is_national_scheme,
        m.group("country"),
        m.group("sp"),
        m.group("reference"),
    )


def assert_parsed_organization_identifier_equal(
    org1: ParsedOrganizationIdentifier,
    org1_source: str,
    org2: ParsedOrganizationIdentifier,
    org2_source: str,
):
    if org1.scheme != org2.scheme:
        raise ValueError(
            f'Mismatched scheme: {org1_source}: "{org1.scheme}", {org2_source}: "{org2.scheme}"'
        )

    if org1.country.casefold() != org2.country.casefold():
        raise ValueError(
            f'Mismatched country: {org1_source}: "{org1.country}", {org2_source}: "{org2.country}"'
        )

    if org1.state_province != org2.state_province:
        org1_state_province = "" if org1.state_province is None else org1.state_province
        org2_state_province = "" if org2.state_province is None else org2.state_province

        raise ValueError(
            f'Mismatched state/province: {org1_source}: "{org1_state_province}", '
            f'{org2_source}: "{org2_state_province}"'
        )

    if org1.reference != org2.reference:
        org1_reference = "" if org1.reference is None else org1.reference
        org2_reference = "" if org2.reference is None else org2.reference

        raise ValueError(
            f'Mismatched registration reference: {org1_source}: "{org1_reference}", '
            f'{org2_source}: "{org2_reference}"'
        )


class OrganizationIdentifierLeiValidator(validation.Validator):
    def __init__(self):
        super().__init__(
            validations=[
                lei.VALIDATION_INVALID_LEI_CHECKSUM,
                lei.VALIDATION_INVALID_LEI_FORMAT,
            ],
            pdu_class=x520_name.X520OrganizationIdentifier,
            predicate=lambda n: any(n.children)
            and str(n.child[1].pdu).startswith("LEI"),
        )

    def validate(self, node):
        value = str(node.child[1].pdu)

        if not value.startswith(LEI_PREFIX):
            # let the syntax validator report this problem
            return

        lei_value = value[len(LEI_PREFIX) :]

        lei.validate_lei(lei_value)


class OrganizationIdentifierElementAllowance(NamedTuple):
    country_codes: Tuple[Set[str], validation.ValidationFinding]
    state_province: Tuple[Rfc2119Word, Optional[str]]
    reference: Tuple[Rfc2119Word, Optional[str]]


class OrganizationIdentifierValidatorBase(validation.Validator):
    def __init__(
        self,
        element_allowances: Dict[str, OrganizationIdentifierElementAllowance],
        invalid_format_validation: Optional[validation.ValidationFinding],
        additional_validations: Optional[List[validation.ValidationFinding]] = None,
        **kwargs,
    ):
        if additional_validations is None:
            additional_validations = []

        self._element_allowances = element_allowances.copy()

        self._invalid_format_validation = invalid_format_validation

        validations = [] + additional_validations

        if self._invalid_format_validation is not None:
            validations.append(self._invalid_format_validation)

        for allowance in element_allowances.values():
            validations.append(allowance.country_codes[1])

            OrganizationIdentifierValidatorBase._create_and_append_validation_finding(
                validations, *allowance.state_province
            )
            OrganizationIdentifierValidatorBase._create_and_append_validation_finding(
                validations, *allowance.reference
            )

        super().__init__(validations=validations, **kwargs)

    @staticmethod
    def _create_and_append_validation_finding(validations, presence_word, finding_code):
        if presence_word != Rfc2119Word.MAY:
            validations.append(
                validation.ValidationFinding(presence_word.to_severity, finding_code)
            )

    @classmethod
    def parse_organization_id_node(
        cls, node: document.PDUNode
    ) -> ParsedOrganizationIdentifier:
        pass

    @classmethod
    def handle_unknown_scheme(
        cls, node: document.PDUNode, parsed: ParsedOrganizationIdentifier
    ):
        pass

    def validate_with_parsed_value(
        self, node: document.PDUNode, parsed: ParsedOrganizationIdentifier
    ):
        scheme_allowance = self._element_allowances.get(parsed.scheme)

        if scheme_allowance is None:
            return self.handle_unknown_scheme(node, parsed)

        allowed_country_codes, finding = scheme_allowance.country_codes

        findings = []
        if parsed.country not in allowed_country_codes:
            findings.append(
                validation.ValidationFindingDescription(
                    finding,
                    f'Invalid country code for registration scheme "{parsed.scheme}": "{parsed.country}"',
                )
            )

        allowance, finding_code = scheme_allowance.state_province

        if parsed.state_province is None and allowance in {
            Rfc2119Word.SHOULD,
            Rfc2119Word.MUST,
        }:
            findings.append(
                validation.ValidationFindingDescription(
                    validation.ValidationFinding(allowance.to_severity, finding_code),
                    f'State/province missing for registration scheme "{parsed.scheme}"',
                )
            )
        elif parsed.state_province is not None and allowance in {
            Rfc2119Word.SHOULD_NOT,
            Rfc2119Word.MUST_NOT,
        }:
            findings.append(
                validation.ValidationFindingDescription(
                    validation.ValidationFinding(allowance.to_severity, finding_code),
                    f'State/province value present for scheme "{parsed.scheme}": "{parsed.state_province}"',
                )
            )

        allowance, finding_code = scheme_allowance.reference

        if parsed.reference is None and allowance in {
            Rfc2119Word.SHOULD,
            Rfc2119Word.MUST,
        }:
            findings.append(
                validation.ValidationFindingDescription(
                    validation.ValidationFinding(allowance.to_severity, finding_code),
                    f'Registration reference missing for scheme "{parsed.scheme}"',
                )
            )
        elif parsed.reference is not None and allowance in {
            Rfc2119Word.SHOULD_NOT,
            Rfc2119Word.MUST_NOT,
        }:
            findings.append(
                validation.ValidationFindingDescription(
                    validation.ValidationFinding(allowance.to_severity, finding_code),
                    f'Registration reference present for scheme "{parsed.scheme}": "{parsed.reference}"',
                )
            )

        return validation.ValidationResult(self, node, findings)

    def validate(self, node):
        try:
            parsed = self.parse_organization_id_node(node)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self._invalid_format_validation, str(e)
            )

        return self.validate_with_parsed_value(node, parsed)



from typing import Optional

from pyasn1.type import univ

from pkilint import document


def get_string_value_from_attribute_node(node: document.PDUNode) -> Optional[str]:
    node = node.children["value"]

    try:
        _, node = node.child
    except ValueError:
        # attribute value has not been decoded
        return None

    # handle DirectoryString CHOICE
    if isinstance(node.pdu, univ.Choice):
        _, node = node.child

    return str(node.pdu)



import ipaddress
import typing

from pyasn1_alt_modules import rfc5280

from pkilint import validation, document
from pkilint.pkix import general_name


class CommonNameValidator(validation.Validator):
    def __init__(
        self,
        allowed_general_name_types: typing.Set[str],
        validation_unknown_value_source: validation.ValidationFinding,
    ):
        super().__init__(
            validations=[validation_unknown_value_source],
            pdu_class=rfc5280.X520CommonName,
        )

        self._allowed_general_name_types = allowed_general_name_types
        self._validation_unknown_value_source = validation_unknown_value_source

    def validate(self, node):
        # unparsed CN, return
        if not any(node.children):
            return

        _, value_node = node.child
        value_str = str(value_node.pdu)

        san_ext_and_idx = node.document.get_extension_by_oid(
            rfc5280.id_ce_subjectAltName
        )

        if san_ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self._validation_unknown_value_source,
                f'Unknown source for value of common name: "{value_str}"',
            )

        san_ext_node, _ = san_ext_and_idx

        try:
            san_value_node = san_ext_node.navigate("extnValue.subjectAltName")
        except document.PDUNavigationFailedError:
            # unparsed SAN extension, return
            return

        for gn in san_value_node.children.values():
            gn_type, gn_value = gn.child

            if gn_type not in self._allowed_general_name_types:
                continue

            if gn_type == general_name.GeneralNameTypeName.DNS_NAME:
                if str(gn_value.pdu) == value_str:
                    return
            elif gn_type == general_name.GeneralNameTypeName.IP_ADDRESS:
                address_octets = gn_value.pdu.asOctets()

                if len(address_octets) == 4:
                    ip_addr = ipaddress.IPv4Address(address_octets)
                elif len(address_octets) == 16:
                    ip_addr = ipaddress.IPv6Address(address_octets)
                else:
                    # Whoa, Nellie! Let the PKIX validator complain about this one
                    continue

                if str(ip_addr) == value_str:
                    return

        raise validation.ValidationFindingEncountered(
            self._validation_unknown_value_source,
            f'Unknown source for value of common name: "{value_str}"',
        )



from importlib.metadata import version
from typing import List

from fastapi import FastAPI, HTTPException

from pkilint.rest import cabf_serverauth, cabf_smime, pkix, etsi, ocsp, crl
from pkilint.rest import model

_PKILINT_VERSION = version("pkilint")
_API_VERSION = "v1.6"

app = FastAPI(
    title="pkilint API", version=_API_VERSION, description="HTTP interface for pkilint"
)

_CERTIFICATE_LINTER_GROUPS = [
    m.create_linter_group_instance()
    for m in (
        pkix,
        cabf_smime,
        cabf_serverauth,
        etsi,
    )
]

_OCSP_PKIX_LINTER = ocsp.create_ocsp_response_linter()
_CRL_PKIX_LINTER = crl.create_crl_linter()


@app.get("/version")
def version() -> model.Version:
    """Retrieves the version of pkilint that this server is using"""
    return model.Version(version=_PKILINT_VERSION)


@app.get("/certificate")
def certificate_linter_groups() -> List[model.LinterGroup]:
    """Retrieves the groups of linters that are available for linting certificates"""
    return _CERTIFICATE_LINTER_GROUPS


def _get_linter_group_by_name(linter_group_name: str):
    try:
        return next(
            (
                l
                for l in _CERTIFICATE_LINTER_GROUPS
                if l.name.casefold() == linter_group_name.casefold()
            )
        )
    except StopIteration:
        raise HTTPException(404, "Linter group with the specified name does not exist")


@app.get("/certificate/{linter_group_name}")
def certificate_linter_group(linter_group_name: str) -> model.LinterGroup:
    """Retrieves the list of linters that are available in the specified linter group"""
    return _get_linter_group_by_name(linter_group_name)


@app.post("/certificate/{linter_group_name}")
def certificate_determine_and_lint(
    linter_group_name: str, doc: model.CertificateModel
) -> model.LintResultListWithLinter:
    """Determines the linter that is most appropriate to lint the specified certificate and then returns the results
    reported by the linter for the certificate"""

    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    parsed_doc = doc.parsed_document

    linter = linter_group_instance.determine_linter(parsed_doc)

    result_list = linter.lint(parsed_doc)

    return model.LintResultListWithLinter(results=result_list.results, linter=linter)


@app.post("/certificate/{linter_group_name}/determine-linter")
def certificate_determine_type(
    linter_group_name: str, doc: model.CertificateModel
) -> model.Linter:
    """Determines the linter that is most appropriate to lint the specified certificate"""
    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    parsed_doc = doc.parsed_document

    return linter_group_instance.determine_linter(parsed_doc)


@app.get("/certificate/{linter_group_name}/{linter_name}")
def linter_validations(
    linter_group_name: str, linter_name: str
) -> List[model.Validation]:
    """Returns the set of validations performed by the specified linter"""
    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    linter_instance = linter_group_instance.get_linter_by_name(linter_name)

    return linter_instance.validations


@app.post("/certificate/{linter_group_name}/{linter_name}")
def certificate_lint(
    linter_group_name: str, linter_name: str, doc: model.CertificateModel
) -> model.LintResultList:
    """Lints the specified certificate with the specified linter"""
    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    linter_instance = linter_group_instance.get_linter_by_name(linter_name)

    parsed_doc = doc.parsed_document

    return linter_instance.lint(parsed_doc)


@app.get("/ocsp/pkix")
def ocsp_linter_validations() -> List[model.Validation]:
    """Returns the set of validations performed by the OCSP response linter"""

    return _OCSP_PKIX_LINTER.validations


@app.post("/ocsp/pkix")
def ocsp_response_lint(doc: model.OcspResponseModel) -> model.LintResultList:
    """Lints the specified OCSP response"""

    parsed_doc = doc.parsed_document

    return _OCSP_PKIX_LINTER.lint(parsed_doc)


@app.get("/crl/pkix/crl")
def crl_linter_validations() -> List[model.Validation]:
    """Returns the set of validations performed by the CRL linter"""

    return _CRL_PKIX_LINTER.validations


@app.post("/crl/pkix/crl")
def crl_lint(doc: model.CrlModel) -> model.LintResultList:
    """Lints the specified CRL"""

    parsed_doc = doc.parsed_document

    return _CRL_PKIX_LINTER.lint(parsed_doc)



from pyasn1.type import univ
from pyasn1.type.constraint import ValueSizeConstraint

from pkilint import document
from pkilint.nist.asn1 import csor


SlhDsaShaTwo128sPublicKeySize = 32
SlhDsaShaTwo128fPublicKeySize = 32

SlhDsaShaTwo192sPublicKeySize = 48
SlhDsaShaTwo192fPublicKeySize = 48

SlhDsaShaTwo256sPublicKeySize = 64
SlhDsaShaTwo256fPublicKeySize = 64

SlhDsaShake128sPublicKeySize = 32
SlhDsaShake128fPublicKeySize = 32

SlhDsaShake192sPublicKeySize = 48
SlhDsaShake192fPublicKeySize = 48

SlhDsaShake256sPublicKeySize = 64
SlhDsaShake256fPublicKeySize = 64


class SlhDsaShaTwo128sPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShaTwo128sPublicKeySize, SlhDsaShake128sPublicKeySize
    )


class SlhDsaShaTwo128fPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShaTwo128fPublicKeySize, SlhDsaShake128fPublicKeySize
    )


class SlhDsaShaTwo192sPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShaTwo192sPublicKeySize, SlhDsaShake192sPublicKeySize
    )


class SlhDsaShaTwo192fPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShaTwo192fPublicKeySize, SlhDsaShake192fPublicKeySize
    )


class SlhDsaShaTwo256sPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShaTwo256sPublicKeySize, SlhDsaShake256sPublicKeySize
    )


class SlhDsaShaTwo256fPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShaTwo256fPublicKeySize, SlhDsaShake256fPublicKeySize
    )


class SlhDsaShake128sPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShake128sPublicKeySize, SlhDsaShake128sPublicKeySize
    )


class SlhDsaShake128fPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShake128fPublicKeySize, SlhDsaShake128fPublicKeySize
    )


class SlhDsaShake192sPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShake192sPublicKeySize, SlhDsaShake192sPublicKeySize
    )


class SlhDsaShake192fPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShake192fPublicKeySize, SlhDsaShake192fPublicKeySize
    )


class SlhDsaShake256sPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShake256sPublicKeySize, SlhDsaShake256sPublicKeySize
    )


class SlhDsaShake256fPublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        SlhDsaShake256fPublicKeySize, SlhDsaShake256fPublicKeySize
    )


ALGORITHM_OID_TO_KEY_MAPPINGS = {
    # pure
    csor.id_slh_dsa_sha2_128f: SlhDsaShaTwo128fPublicKey(),
    csor.id_slh_dsa_sha2_128s: SlhDsaShaTwo128sPublicKey(),
    csor.id_slh_dsa_sha2_192f: SlhDsaShaTwo192fPublicKey(),
    csor.id_slh_dsa_sha2_192s: SlhDsaShaTwo192sPublicKey(),
    csor.id_slh_dsa_sha2_256f: SlhDsaShaTwo256fPublicKey(),
    csor.id_slh_dsa_sha2_256s: SlhDsaShaTwo256sPublicKey(),
    csor.id_slh_dsa_shake_128f: SlhDsaShake128fPublicKey(),
    csor.id_slh_dsa_shake_128s: SlhDsaShake128sPublicKey(),
    csor.id_slh_dsa_shake_192f: SlhDsaShake192fPublicKey(),
    csor.id_slh_dsa_shake_192s: SlhDsaShake192sPublicKey(),
    csor.id_slh_dsa_shake_256f: SlhDsaShake256fPublicKey(),
    csor.id_slh_dsa_shake_256s: SlhDsaShake256sPublicKey(),
    # pre-hashed
    csor.id_hash_slh_dsa_sha2_128f_with_sha256: SlhDsaShaTwo128fPublicKey(),
    csor.id_hash_slh_dsa_sha2_128s_with_sha256: SlhDsaShaTwo128sPublicKey(),
    csor.id_hash_slh_dsa_sha2_192f_with_sha512: SlhDsaShaTwo192fPublicKey(),
    csor.id_hash_slh_dsa_sha2_192s_with_sha512: SlhDsaShaTwo192sPublicKey(),
    csor.id_hash_slh_dsa_sha2_256f_with_sha512: SlhDsaShaTwo256fPublicKey(),
    csor.id_hash_slh_dsa_sha2_256s_with_sha512: SlhDsaShaTwo256sPublicKey(),
    csor.id_hash_slh_dsa_shake_128f_with_shake128: SlhDsaShake128fPublicKey(),
    csor.id_hash_slh_dsa_shake_128s_with_shake128: SlhDsaShake128sPublicKey(),
    csor.id_hash_slh_dsa_shake_192f_with_shake256: SlhDsaShake192fPublicKey(),
    csor.id_hash_slh_dsa_shake_192s_with_shake256: SlhDsaShake192sPublicKey(),
    csor.id_hash_slh_dsa_shake_256f_with_shake256: SlhDsaShake256fPublicKey(),
    csor.id_hash_slh_dsa_shake_256s_with_shake256: SlhDsaShake256sPublicKey(),
}

ALGORITHM_OID_TO_PARAMETER_MAPPINGS = {
    k: document.ValueDecoder.VALUE_NODE_ABSENT
    for k in ALGORITHM_OID_TO_KEY_MAPPINGS.keys()
}
